from typing import List


from madas.backend.backend_core import Backend
from madas.material import Material
from madas.utils import safe_log, JSONNumpyEncoder

class DictBackend(Backend):
    """
    Storage of materials data in a dictionary in memory.

    Inteded use are development and testing, however, data can be stored by `json.dump` ing the dict.
    """

    def __init__(self, 
                 filename=None, 
                 filepath=None, 
                 make_dirs=True, 
                 key_name="mid", 
                 log=None):
        self.key_name = key_name
        self._metadata = {}
        self._dict = {}
        self.set_logger(log)
        #  kept for compatibility, but ignored in the remainder
        self.filename = filename 
        self.filepath = filepath

    @property
    def abs_path(self):
        """
        Included for compatibility, returns `None`.
        """
        return None

    def get_single(self, mid: str) -> Material:
        """
        Get a single entry from the database.
        """
        return Material.from_dict(self._dict[mid])

    def get_many(self, mids: List[str]) -> List[Material]:
        """
        Get several entry from the database.
        """
        return [self.get_single(mid) for mid in mids]

    def get_by_id(self, db_id: int) -> Material:
        """
        Return a single entry from an (integer valued) database id.
        """
        mid = list(self._dict.keys())[db_id]
        return self.get_single(mid)

    def add_single(self, material: Material) -> None:
        """
        Add data to the database.
        """
        self._dict.update(**{material.mid : material.to_dict()})
        safe_log(self._log_write_message(material.mid), logger=self.log, level = "info")
        
    def add_many(self, materials: List[Material]) -> None:
        """
        Add data to the database.
        """
        new_data = {material.mid : material.to_dict() for material in materials}
        self._dict.update(**new_data)

    def update_single(self, mid: str, **kwargs) -> None:
        """
        Update a single entry in the database.

        The `properties` attribute of entry `mid` will be updated with `**kwargs`.
        """
        self._dict[mid]["properties"].update(**kwargs)
    
    def update_many(self, mids: List[str], kwargs_list: List[dict] = []):
        """
        Update several entries in the database.

        The `properties` attributes of entries `mids` will be updated with entries of `kwargs_list`.
        """
        for mid, props in zip(mids, kwargs_list):
            self._dict[mid]["properties"].update(**props)

    def update_metadata(self, **kwargs) -> None:
        """
        Updata database metadata.
        """
        self._metadata.update(**kwargs)

    def has_entry(self, entry_id: str) -> bool:
        """
        Check if an entry with the given id is present in the database.
        """
        if entry_id in self._dict.keys():
            return True
        return False

    def get_length(self) -> int:
        """
        Return the length of the database, i.e. the total number of entries.
        """
        return len(self._dict)

    def to_json(self) -> str:
        """
        `json.dumps` the content of the database.
        """
        import json
        print(self._dict)
        return json.dumps({"data": self._dict, "metadata" : self.metadata}, cls=JSONNumpyEncoder)

    @staticmethod
    def from_json(json_data) -> object:
        """
        Restore data from string. 
        Inverse option to `DictBackend().to_json()`.
        """
        import json
        data = json.loads(json_data)
        self = DictBackend()
        self._dict = data["data"]
        self._metadata = data["metadata"]
        return self


from madas import Fingerprint, Material
import pytest

@pytest.fixture
def fingerprint():
    return Fingerprint()

@pytest.fixture()
def test_material():
    return Material("a:b", atoms=None, data = {"test" : "data"})

def test_imports():
    
    from madas.fingerprints.DUMMY_fingerprint import DUMMYFingerprint, DUMMY_similarity
    from madas.fingerprint import import_builtin_module
    
    fingerprint_class, similarity_function = import_builtin_module("DUMMY")

    assert fingerprint_class == DUMMYFingerprint, "Error importing fingerprint class"
    assert similarity_function == DUMMY_similarity, "Error importing similarity function"

def test_specify(fingerprint):

    from madas.fingerprints.DUMMY_fingerprint import DUMMYFingerprint

    def mock_target_function(x: list):
        return sum(x)

    fingerprint.specify(DUMMYFingerprint, target_function = mock_target_function)

    assert isinstance(fingerprint, DUMMYFingerprint), "Fingerprint class was not updated during `specify`"

    fingerprint.set_data("data", [1,2,3])

    assert fingerprint.y == 6, "Target function was not passed as kwarg to fingerprint"

def test_specify_on_init():

    from madas.fingerprints.DUMMY_fingerprint import DUMMYFingerprint

    def mock_target_function(x: list):
        return sum(x)

    def mock_similarity_function(*args, **kwargs):
        return -1

    fingerprint = Fingerprint("DUMMY", target_function = mock_target_function, similarity_function=mock_similarity_function)

    assert isinstance(fingerprint, DUMMYFingerprint), "Specify was not called upon initialization."

    fingerprint.set_data("data", [1,2,3])

    assert fingerprint.y == 6, "Target function was not passed as kwarg to fingerprint"

    assert fingerprint.get_similarity(fingerprint) == -1, "Did not use similarity function provided upon initialization."

    fingerprint = Fingerprint("DUMMY", target_function = mock_target_function).calculate([1,2,3])

    assert fingerprint.get_similarity(fingerprint) == 1.0, "Did not use specified similarity function."

    class NewFingerprint(Fingerprint):

        def __init__(self, a = 1, b = 1):
            self.a = a
            self.b = b

        def calculate(self, *args):
            self.set_data("result", self.a + self.b)
            return self

    fingerprint = Fingerprint(NewFingerprint, a=2, b=3)

    assert fingerprint.calculate().data["result"] == 5, "Did not initialize fingerprint from type object"
    assert fingerprint.fp_type == "NewFingerprint", "Did not initialize fingerprint from type object"

def test_subclassing(test_material):

    class SubFingerprint(Fingerprint):

        def __init__(self, a = 1, b = 1):
            self.a = a
            self.b = b

        def calculate(self, material):
            self.set_mid(material)
            self.set_data("result", self.a + self.b)
            return self

    sub_fp = SubFingerprint().calculate(test_material)

    assert sub_fp.mid == test_material.mid, "Did not set mid from material in subclass"


def test_serialize(test_material):

    class SubFingerprint(Fingerprint):

        def __init__(self, a = 1, b = 1):
            self.a = a
            self.b = b

        def calculate(self, material):
            self.set_mid(material)
            self.set_data("result", self.a + self.b)
            return self

    sub_fp = SubFingerprint().calculate(test_material)

    data = sub_fp.serialize()

    sub_fp2 = Fingerprint.deserialize(data)

    assert sub_fp == sub_fp2, "Could not deserialize fingerprint correctly."


def test_set_data(fingerprint):

    fingerprint.set_data("test", "data")

    assert fingerprint.data["test"] == "data", "Data was not set correctly."

    assert fingerprint["test"] == "data", "Data could not be retrieved via __getitem__"

    fingerprint.set_data("list", [1,2,3])

    assert fingerprint["list"] == [1,2,3], "Failed to set two different data entries"

def test_set_similarity(fingerprint):

    def dummy_similarity(*args, trigger=False):
        if trigger:
            return "Passed with kwargs"
        else:
            return "Passed"

    fingerprint.set_similarity_function(dummy_similarity)

    assert fingerprint.get_similarity(Fingerprint()) == "Passed", "Unable to set similarity function"

    fingerprint.set_similarity_function(dummy_similarity, trigger=True)

    assert fingerprint.get_similarity(Fingerprint()) == "Passed with kwargs", "Unable to set similarity function with kwargs"

def test_get_similarity(fingerprint):

    fp2 = Fingerprint()

    def mock_similarity_function(*args):
        return True

    fingerprint.set_similarity_function(mock_similarity_function)

    assert all(fingerprint.get_similarities([fp2, fp2])), "Could not calculate similarities for several fingerprints"

    fingerprint.set_pass_on_exceptions(False)

    fp2.set_fp_type("Whatever")

    with pytest.raises(TypeError):
        # This should fail because the fingerprint types are not the same
        fingerprint.get_similarity(fp2)

    with pytest.raises(TypeError):
        # This should fail because the other is no fingerprint
        fingerprint.get_similarity(0)

    fingerprint.set_pass_on_exceptions(True)

    assert fingerprint.get_similarity(fp2) == 0, "Passing on exceptions, invalid fingerprints should return 0 as similarity"

    assert fingerprint.get_similarity(0) == 0, "Passing on exceptions, invalid fingerprints should return 0 as similarity"

    def mock_failing_similarity_function(*args):
        raise AssertionError("I am failing.")

    fingerprint.set_similarity_function(mock_failing_similarity_function)

    fp2 = Fingerprint()

    assert fingerprint.get_similarity(fp2) == 0, "Passing on exceptions, failing similarity functions should return 0 as similarity"

    fingerprint.set_pass_on_exceptions(False)

    with pytest.raises(AssertionError):
        # This should fail because the similarity function raises an exception
        fingerprint.get_similarity(fp2)

@pytest.mark.filterwarnings("error")
def test_name_mismatch_warning():

    def mock_similarity(a,b):
        return 0
    
    a = Fingerprint(name="a", similarity_function=mock_similarity)
    b = Fingerprint(name="b", similarity_function=mock_similarity)

    with pytest.warns(UserWarning):
        a.get_similarity(b)

    # test that the warning is not emitted if the names are the same  
    a.get_similarity(a)
    b.get_similarity(b)
    


import logging
import sys
import traceback
from typing import Callable, List
from typing import Any
from hashlib import md5

from madas.material import Material
from madas.utils import safe_log

class APIClass():
    """
    Base class for APIs to different data sources.
    """

    def __init__(self, 
                 logger = None):
        self.set_logger(logger)

    def get_calculation(self, *args, **kwargs) -> Material:
        """
        Get a single material from the external data source.

        **Expected input:**

        ids: *Any*
            Any piece of data that allows the external resource to identify which data shall be returned.

        **Expected to return:**

        material: *madas.Material*
            Material data in a Material object.

        **Raises:**

        NotImplementedError: This function is not implemented here.
        """
        raise NotImplementedError('The option to get individual calculations is not implemented in this API.')
        return Material()

    def get_calculations_by_search(self, *args, **kwargs) -> List[Material]:
        """
        Get a list of materials from the external data source.

        **Expected input:**

        search_query: *Any*
            Any piece of data that allows the external resource to identify which data shall be returned.

        **Expected to return:**

        material: *List[madas.Material]*
            List of Material objects, containing the data from external source.

        **Raises:**

        NotImplementedError: This function is not implemented here.
        """
        raise NotImplementedError('The option to get sets of calculations is not implemented in this API.')
        return [Material()]

    def get_property(self, **kwargs) -> Any:
        """
        Get a property from the external source.

        **Expected input:**

        kwargs: *Any*
            Any piece of data that allows the external resource to identify which property for which data shall be returned.

        **Expected to return:**

        property: *Any*
            A single property, e.g. a string or float, that can be stored to the database.

        **Raises:**

        NotImplementedError: This function is not implemented here.
        """
        raise NotImplementedError('The option to get individual properties is not implemented in this API.')
        return {'property':None}
    
    def hash_query(self, *args, **kwargs) -> str:
        """
        Hash a query that is passed to the API. This is used to avoid repeatedly querying the same data.
        """
        query_string = ""
        for arg in sorted(map(str, args)):
            query_string+=arg
        for key, val in sorted(kwargs.items(), key = lambda x: x[0]):
            query_string+=f"{key}{val}"
        return md5(query_string.encode()).hexdigest()

    def _gen_mid(self, *args) -> str:
        """
        Generate the unique material identifier from arguments. Mids are used to identify the material both in the external database and the *MaterialsDatabase*.

        **Expected input:**

        args: *Any*
            Any piece of data that allows to generate a unique id.

        **Expected to return:**

        mid: *str*
            Unique identifier of the material.

        **Raises:**

        NotImplementedError: This function is not implemented here.        
        """
        raise NotImplementedError('The option to generate material ids is not implemented in this API.')
        return '<unique>:<identifyer>'

    def set_logger(self, logger: logging.Logger) -> None:
        """
        Set logger of the API. 
        The log is used to improve analysis of errors during collection of materials and increase robustness due to failures of external resources.

        **Arguments:**

        logger: *logging.Logger* or *None*
            Logger object.
            If *logger = None*, errors will be written to STDERR.

        **Returns:**

        None
        """
        self.log = logger

    def _report_error(self, error_message, level = "error"):
        """
        Write error to logging.log (if exists), or stderr.
        """
        safe_log(error_message, logger = self.log, level=level)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.__dict__})"

class APIError(Exception):
    """
    A generic exception to specify errors that are raised in an *APIClass*.

    **Usage:**

        message = "Something went wrong!"

        raise APIError(message)
    """

    def __init__(self, message: str):
        super().__init__(message)

class DatabaseEntryDoesNotExistError(Exception):
    """
    A generic exception to be raised if a specific entry does not exist in an external database.

    **Usage:**

        <code that can not retieve data>

        raise DatabaseEntryDoesNotExistError("Can not retrieve entry!")
    """
    def __init__(self, message):
        super().__init__(message)

def api_call(call, 
             retries: int = 3, 
             report_function: Callable = print, 
             report_function_parameters: dict = {"file" : sys.stderr}) -> Callable:
    """
    Decorator function to realize 10 tries for an API call before raising an exception.
    For each failed function call, the decorator will pass the exception name and the number of trials to *report_function*.
    If the number of trials is exceeded, the decorator will pass the name of the failing function, the arguments and keyword arguments to *report_function*.

    **Arguments:**
    
    call: *Callable*
        Function that tries to make an API call to an external database.

    **Keyword arguments:**

    retries: `int`
        Number of retries before passing on the exceptions

        Default: `3`

    report_function: *Callable*
        Function to report error.

        Default: print

    report_function_parameters: `dict`
        Keyword arguments passed to report function.
        
        default: ``{"file" : sys.stderr}`` 

    **Raises:**

    APIError: could not retrieve data

    """
    def call_api(*args, **kwargs):
        trials = 0
        success = False
        while trials < retries and not success:
            try:
                answer = call(*args, **kwargs)
                success = True
            except Exception as error:
                error_message = f'Failed connection to server because of {type(error).__name__}: {str(error)}\nThis was attempt number {str(trials)}.\ntraceback: {traceback.format_exc()}'
                report_function(error_message, **report_function_parameters)
                trials += 1
        if not success:
            error_message = f'Function {call.__name__}: Could not connect to external API! Using args: {args} and kwargs: {kwargs}.'
            report_function(error_message, **report_function_parameters)
            raise APIError(error_message)
        return answer

    return call_api



from nomad_dos_fingerprints.DOSfingerprint import DOSFingerprint # noqa: F401
from nomad_dos_fingerprints.grid import Grid # noqa: F401
from nomad_dos_fingerprints.similarity import tanimoto_similarity # noqa: F401


from typing import Callable 
from typing import Any

from madas import Material
from madas.fingerprint import Fingerprint
from madas.utils import report_error, resolve_nested_dict

def PROP_similarity(fingerprint1: Fingerprint, 
                    fingerprint2: Fingerprint, 
                    delta: float = 1, 
                    scaling_function: Callable = lambda x: x) -> float:
    """
    Calculate similarity between PROPFingerprints.

    Calculates *S(a, b) = delta / ( scaling_function(|a - b|) + delta )*

    where `S` is the similarity, `a` and `b` are scalar material properties encoded in `PROPFingerprint` objects, `delta` is an arbitrary scalar value to rescale the similarity, and `scaling_function` is any function f(x) -> y with real numbers x,y >= 0.

    **Arguments:**

    fingerprint1, fingerprint2: `Fingerprint`
        PROPFingerprints of to calculate similarity

    **Keyword arguments:**

    delta: `float`
        Scaling factor that allows to define how the absolute differences in the scalar property affect the similarity.

    scaling_function: `Callable`
        Function to dynamically rescale the absolute differences in properties.

    **Returns:**

    similarity: `float`
        Similarity between input fingerprints.
    """
    if not fingerprint1.fp_type == fingerprint2.fp_type == "PROP":
        raise TypeError("PROP similarity can only be used for PROP fingerprints.")
    property_name = fingerprint1.data["property_path"]
    if not fingerprint2.data["property_path"] == property_name:
        raise ValueError("Trying to calculate similarity of different properties, which is impossible.")
    value1 = fingerprint1.data["value"]
    value2 = fingerprint2.data["value"]
    distance = abs(value1 - value2)
    similarity = delta / (scaling_function(distance) + delta)
    return similarity

class PROPFingerprint(Fingerprint):
    """
    A fingerprint object to represent a user defined property.

    Provide the path to a property to compute 
    """

    def __init__(self, 
                 name = "PROP", 
                 property_path: str | Callable = 'atomic_density',
                 similarity_function = PROP_similarity, 
                 from_data: bool = True):
        self._from_data = from_data
        if callable(property_path):
            self.set_data("property_path", f"function_{property_path.__name__}")
            self._compute_function = property_path
        else:      
            self.set_data("property_path", property_path)
            self._compute_function = None
        self.set_fp_type("PROP")
        self.set_name(name)
        self.set_similarity_function(similarity_function)

    def from_material(self, material: Material) -> object:
        """
        Set fingerprint data to property_path given during initialization. 

        Defines the following data in `Fingerprint.data`:

        *value* : Poperty value, extracted from Material.data, specified by *property_path*.

        **Returns:**

        self: *PROPFingerprint*

        **Raises:**

        KeyError: if ``self.pass_on_exceptions == False`` and the material has not data entry with `property_path`.
        """
        self.set_mid(material)
        if self._compute_function is not None:
            self.set_data("value", self._compute_function(material))
            return self
        try:
            prop = material.get_data_by_path(self.data["property_path"])
        except KeyError as e:
            message = f'No property of type {self.data["property_path"]} for material {material.mid}.'
            if self.pass_on_exceptions:
                report_error(None, message)
                return None
            else:
                report_error(None, message)
                raise e
        self.set_data("value", prop)
        return self

    def calculate(self, data: Any = None) -> object:
        """
        Added for compatibility with the API.
        """
        if self._compute_function is not None:
            self.set_data("value", self._compute_function(data))
            return self
        try:
            prop = resolve_nested_dict(data, self.data['property_path'], fail_on_key_error=True)
        except KeyError as key_error:
            message = f'No property of type {self.data["property_path"]}.'
            if self.pass_on_exceptions:
                report_error(None, message)
                return None
            else:
                report_error(None, message)
                raise key_error
        self.set_data("value", prop)
        return self



import pytest

from madas import Material

@pytest.fixture()
def material():
    return Material("a")

def test_init(material):
    assert material.mid == "a", "Did not set mid correctly"
    assert material.atoms is None, "Did not set atoms correctly"
    assert material.data == {}, "Did not set data correctly"
    assert material.properties == {}, "Did not set properties correctly"

def test_get_property_by_path(material):

    with pytest.raises(KeyError):
        material.get_property_by_path("test/data")

    material.set_properties({"test" : {"data" : 1}})

    assert material.get_property_by_path("test/data") == 1, "Did not retrieve property by path"

def test_get_data_by_path(material):

    with pytest.raises(KeyError):
        material.get_data_by_path("test/data")

    material.set_data({"test" : {"data" : 1}})

    assert material.get_data_by_path("test/data") == 1, "Did not retrieve data by path"


import pytest

import os

import pandas as pd

from ase.build import bulk
from ase.io import write

from madas.apis.local_data_API import FileReaderASE, CSVPropertyReader, API
from madas import Material, MaterialsDatabase

@pytest.fixture()
def atoms_test_data(tmpdir):
    atoms = bulk("Si")
    for folder_name in ["a", "b"]:
        os.makedirs(os.path.join(tmpdir, folder_name))
        filepath = os.path.join(tmpdir, folder_name, "geometry.in")
        write(filepath, atoms)
    return tmpdir, atoms

@pytest.fixture()
def csv_test_data(tmpdir):
    df = pd.DataFrame([[1, 2 ], [3, 4]], index = ["a", "b"], columns = ["property1", "property2"])
    filepath = os.path.join(tmpdir, "properties.csv")
    df.to_csv(filepath)
    return tmpdir, df.transpose().to_dict()

def test_FileReaderASE(atoms_test_data):

    tmpdir, atoms = atoms_test_data

    data_dir = os.path.join(tmpdir, "a")

    reader = FileReaderASE(file_path=data_dir)

    read_atoms = reader.read("geometry.in")

    assert read_atoms == atoms, "Atoms from file are not identical with original"

def test_CSVPropertyReader(csv_test_data):

    dir_, data = csv_test_data

    reader = CSVPropertyReader(dir_)

    read_data = reader.read("properties.csv")

    assert data == read_data, "Did not read correct data from CSV file"

def test_API_get_calculation(atoms_test_data, csv_test_data):

    tmpdir, atoms = atoms_test_data
    _, data = csv_test_data

    api = API(root=tmpdir)

    ref_material = Material(mid = api._gen_mid("a", "geometry.in", ".", "properties.csv"),
                            atoms=atoms,
                            data=data["a"])

    mat = api.get_calculation("a", "geometry.in", ".", "properties.csv", property_file_id="a")

    assert ref_material == mat, "Did not read correct material from test data"

def test_API_get_calculations_by_search(atoms_test_data, csv_test_data):
    
    tmpdir, atoms = atoms_test_data
    _, data = csv_test_data

    api = API(root=tmpdir)

    ref_material_a = Material(mid = api._gen_mid(os.path.join(tmpdir, "a"), 
                                                "geometry.in", 
                                                ".", 
                                                "properties.csv"),
                            atoms=atoms,
                            data=data["a"])
    ref_material_b = Material(mid = api._gen_mid(os.path.join(tmpdir, "b"), 
                                                "geometry.in", 
                                                ".", 
                                                "properties.csv"),
                            atoms=atoms,
                            data=data["b"])

    read_data = api.get_calculations_by_search(tmpdir, "geometry.in", ".", "properties.csv")

    assert len(read_data) == 2, "wrong number of data points read"
    assert ref_material_a in read_data, "did not read material a"
    assert ref_material_b in read_data, "did not read material b"

def test_API_get_materials_from_different_locations(tmpdir, monkeypatch):

    def mock_calculations_by_search(folder_path: str, 
                                   file_name: str, 
                                   property_file_path: str, 
                                   property_file_name: str, 
                                   file_reader_kwargs: str = {}, 
                                   property_reader_kwargs: str = {}):
        if folder_path == "1":
            return [Material("a")]
        else:
            return [Material("b")]

    monkeypatch.setattr(API, "get_calculations_by_search", mock_calculations_by_search)

    db = MaterialsDatabase(filepath=tmpdir, api=API(), log_mode="stream")

    db.fill_database("1", None, None)

    db.fill_database("2", None, None)


from typing import Any
import numpy as np
import sys
from json import JSONEncoder
import logging 

# check which progress bar to use, from: https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook
is_jupyter = False
try:
    shell = get_ipython().__class__.__name__  # type:ignore
    if shell == 'ZMQInteractiveShell':
        is_jupyter = True   # Jupyter notebook or qtconsole
    elif shell == 'TerminalInteractiveShell':
        is_jupyter = False  # Terminal running IPython
    else:
        is_jupyter = False  # Other type (?)
except NameError:
    is_jupyter = False      # Probably standard Python interpreter

if is_jupyter:
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm  # noqa: F401

def rmsle(y_true, y_pred):
    """
    Root means square logarithmic error as used in the NOMAD kaggle competition:

    Sutton, C., Ghiringhelli, L.M., Yamamoto, T. et al. 
    Crowd-sourcing materials-science challenges with the NOMAD 2018 Kaggle competition. 
    npj Comput Mater 5, 111 (2019). https://doi.org/10.1038/s41524-019-0239-3
    
    **Arguments:**

    y_true: `List[float]`
        List of true target values

    y_pred: `List[float]`
        List of predicted target values

    **Returns:**

    rmsle: `float`
        Root mean squared logarithmic error
    """
    if not isinstance(y_true, (list, np.ndarray)):
        y_true = [y_true]
    if not isinstance(y_pred, (list, np.ndarray)):
        y_pred = [y_pred]
    errors = [np.log((np.array(y_p) + 1)/(np.array(y_t) + 1))**2 for y_t, y_p in zip(y_true, y_pred)]
    msle = sum(errors) / len(errors)
    return np.sqrt(msle).tolist()

def report_error(logger: logging.Logger, error_message: str):
    """
    Report error by writing it to a `logging.Logger` instance or to `sys.stderr`.

    **Arguments:**

    logger: `logging.Logger` or `None`
        Log target. Write to log or to stderr if `logger == None`
    
    error_message: `str`
        Message to write

    **Returns:**

    `None`
    """
    if logger is not None:
        logger.error(error_message)
    else:
        print(error_message, file=sys.stderr)

def safe_log(message: str,
             logger: logging.Logger = None, 
             level: str = "error"):    
    """
    Report error by writing it to a `logging.Logger` instance or to `sys.stderr`.

    **Arguments:**
    
    error_message: `str`
        Message to write

    **Keyword arguments:**

    logger: `logging.Logger` or `None`
        Log target. Write to log or to stderr if `logger == None`

        default: `None`

    level: `str`
        Choose target of log. Write to logger.info, logger.error, or, stdout, stderr.

        Options: "error", "info"

        default, and fallback, is "error"

    **Returns:**

    `None`
    """
    if logger is not None:
        if level == "info":
            logger.info(message)
        else:
            logger.error(message)
    else:
        out = sys.stdout if level == "info" else sys.stderr
        print(message, file = out)

def seed_random_number_generators(random_seed):
    """
    Seed Python standard library random generator and numpy random generator.

    **Arguments:**

    random_seed: `Any`
        Seed passed to random number generators.
    """
    from random import seed
    seed(random_seed)
    np.random.seed(random_seed)

def resolve_nested_dict(archive: dict, 
                        path:str, 
                        error_message="failed to resolve path", 
                        fail_on_key_error=False) -> Any:
    """
    Given a nested dictionary (including lists), return the entry at `path` in the dictionary.
    
    **Arguments:**
    
    archive: *Dict[Dict, List]*
        Nested dictionary with str or int keys
        
        Example: `{'a' : {'b' : [{'c':5}]}, 'd' : 'data'}`
        
    path: *str*
        Keys to navigate the dictionary given as `archive`, separated by a '/'
    
        Example: `'a/b/0/c'` # return value `5`
        
    **Keyword arguments:**
    
    error_message: *str*
        Message to display if the path can not be resolved
        
        default: `"failed to resolve path"`
        
    fail_on_key_error: *bool*
        If the path can not be resolved, raise the Exception insted of returning `None`.
        
    **Returns:**
    
    archive: *Any*
        Value at specified location in the dictionary.
        `None`, if the path can not resolved.
        
    **Raises:**
    
    `KeyError`
        Path can not be resolved and `fail_on_key_error==True`
    """
    try:
        for key in path.strip("/").split("/"):
            try:
                key_int = int(key)    
                archive = archive[key_int]
            except ValueError:
                archive = archive[key]
        return archive
    except KeyError as e:
        print(error_message, file=sys.stderr)
        if fail_on_key_error:
            raise e
        else:
            return None

def list_chunks(long_list, chunk_length = 5):
    chunked_list = []
    chunk = []
    for index,item in enumerate(long_list):
        chunk.append(item)
        if index % chunk_length == 0:
            chunked_list.append(chunk)
            chunk = []
    if chunk != []:
        chunked_list.append(chunk)
    return chunked_list

def merge_k_nearest_neighbor_dicts(fp_type_list, dicts):
    merged_dict = {}
    for index, fp_type in enumerate(fp_type_list):
        materials = [key for key in dicts[index].keys()]
        for material in materials:
            data = dicts[index][material]
            if material not in merged_dict.keys():
                merged_dict[material] = {mid:{fp_type:sim} for mid, sim in data}
            else:
                for mid, sim in data:
                    if mid not in merged_dict[material].keys():
                        merged_dict[material][mid] = {fp_type:sim}
                    else:
                        merged_dict[material][mid].update({fp_type:sim})
    return merged_dict

def contour_data_from_list(coord_list):
    xs = np.unique([x[0] for x in coord_list])
    ys = np.unique([x[1] for x in coord_list])
    z_matrix = []
    for y in ys:
        row = []
        for x in xs:
            results = [entry for entry in coord_list if (entry[0] == x).all() and (entry[1] == y).all()]
            if len(results) > 1:
                print(results)
                raise AssertionError('Double counting of coordinates.')
            row.append(results[0][2])
        z_matrix.append(row)
    return np.array(z_matrix), xs, ys

class JSONNumpyEncoder(JSONEncoder):
    """
    Encoder to 
    """
    def default(self, obj: Any) -> Any:
        if isinstance(obj, np.ndarray):
            obj=obj.tolist()
        if isinstance(obj,np.floating):
            obj=float(obj)
        return obj

class BatchIterator():
    """
    A helper class for paralellization of calculation of large batched similarity matrices.

    A `batch` is defined here as a nested list of integers describing the indices of a sub-matrix.

    To parallelize the calculation of large matrices of different nodes in a compute cluster, 
    each node may compute several sub-matrices that can be combined to the complete matrix.

    To reduce the memory footprint, each node loads only a subset of the required fingerprints
    and calculates the (overlap) similarity matrix from these.

    This class iteratively provides the mapping of fingerprint indices in a given list.
    Thus, upon iteration over an object of this class, all combinations of indices that are 
    required to compute a unique sub-matrix for a specific task id are returned.
    """
    
    def __init__(self, 
                 n_entries: int, 
                 batch_size: int, 
                 n_tasks: int = 1, 
                 task_id: int = 0,
                 symmetric: bool = True):
        """
        **Arguments:**

        n_entries: `int`
            Total number of indices to consider
        
        batch_size: `int`
            Range of indices for each batch

        **Keyword arguments:**

        n_tasks: `int`
            Total number of tasks executed in parallel

            default: `1`

        task_id: `int`
            Integer number that specifies which tasks this instance is doing

            Must be smaller than `n_tasks`

            default: `0`

        symmetric: `bool`
            Return only symmetrically unique batches

            default: `True`
        """
        if task_id >= n_tasks:
            raise ValueError("Task ID must be smaller than number of tasks!")
        self.n_tasks = n_tasks
        self.task_id = task_id
        self.symmetric = symmetric
        self._n_entries = n_entries
        self._batch_size = batch_size
        self._batches = self._gen_batches(n_entries, batch_size)
        self._iter_index = 0

    @property
    def batches(self):
        """
        _All_ batches, regardless of the task id.
        """
        return self._batches
    
    @property
    def n_entries(self):
        """
        Total number of entries, i.e. the size of the range of integer to consider.
        """
        return self._n_entries
    
    @property
    def batch_size(self):
        """
        Range of integers in a batch.
        """
        return self._batch_size
    
    def plot_batches(self, figure = True, show = True, text_fontsize=10):
        """
        Generate a plot of the batches for visualization.

        **Keyword arguments:**

        figure: `bool`
            Create new matplotlib.pyplot.figure

            default: `True`

        show: `bool`
            show plot

            default: `True`
        """
        import matplotlib.pyplot as plt
        if figure:
            plt.figure(figsize = (10,10))
        plt.xlim(0,self.n_entries)
        plt.ylim(self.n_entries,0)
        for batch in self:
            plt.fill_between([batch[0][0], batch[0][1]],[batch[1][1], batch[1][1]], [batch[1][0], batch[1][0]], color = plt.get_cmap("cool")(self.task_id / self.n_tasks))
            plt.text(sum(batch[0])/2, sum(batch[1])/2, f"{batch}\n{self.task_id}", horizontalalignment='center', fontsize=text_fontsize)
        if show:
            plt.show()
            
    def plot_batch_rows(self, figure = True, show = True, text_fontsize=10):
        """
        Generate a plot of the batches in a row for visualization.

        **Keyword arguments:**

        figure: `bool`
            Create new matplotlib.pyplot.figure

            default: `True`

        show: `bool`
            show plot

            default: `True`
        """
        import matplotlib.pyplot as plt
        if figure:
            plt.figure(figsize = (10,10))
        plt.xlim(0,self.n_entries)
        plt.ylim(self.n_entries,0)
        for idx, batch_row in enumerate(self.get_batch_rows()):
            for batch in batch_row:
                plt.fill_between([batch[0][0], batch[0][1]],[batch[1][1], batch[1][1]], [batch[1][0], batch[1][0]], color = plt.get_cmap("tab20")(idx), alpha = 0.5)
                plt.text(sum(batch[0])/2, sum(batch[1])/2, f"{batch}\n{self.task_id}", horizontalalignment='center', fontsize=text_fontsize)
        if show:
            plt.show()
    
    def get_batches_for_index(self, index: int):
        """
        Return all batches that contain a given index.

        **Arguments:**

        index: `int`
            Index to that should be contained in returned batches.

        **Returns:**

        batches: `List[List[List[int]]]`
            batches containing the given index
        """
        batches = []
        for batch in self.batches:
            if self.symmetric:
                if batch[0][0] <= index < batch[0][1] or batch[1][0] <= index < batch[1][1]:
                    batches.append(batch)
            else:
                if batch[1][0] <= index < batch[1][1]:
                    batches.append(batch)
        return batches
    
    def get_batch_rows(self):
        """
        This is sensitive to the task index!
        Get all rows of batches.
        """
        batch_row_start_indices = list(range(0, self.n_entries, self.batch_size))
        batch_rows_for_task_id = []
        for idx, index in enumerate(batch_row_start_indices):
            if idx % self.n_tasks == self.task_id:
                batch_rows_for_task_id.append(self.get_batches_for_index(index))
        return batch_rows_for_task_id
    
    def _gen_batches(self, size, batch_size):
        batch_list = []
        batch_x_list = self.linear_batch_list(size, batch_size)
        if self.symmetric:
            for idx, batch_x in enumerate(batch_x_list):
                for batch_y in batch_x_list[idx:]:
                    batch_list.append([batch_x, batch_y])
        else:
            for batch_x in batch_x_list:
                for batch_y in batch_x_list:
                    batch_list.append([batch_x, batch_y])            
        return batch_list
    
    @staticmethod
    def linear_batch_list(size, batch_size):
        """
        List of indices that splits a list of size `size` into lists of length `batch_size`.
        """
        len_batched = int(size / batch_size)
        if len_batched * batch_size < size:
            len_batched += 1
        batch_x_list = []
        batch_index = 0
        for _ in range(len_batched):
            if batch_index + batch_size > size:
                batch_x_list.append([batch_index, size])
                break
            else:
                batch_x_list.append([batch_index, batch_index + batch_size])
                batch_index += batch_size
        return batch_x_list
    
    def __len__(self):
        return len(self._batches)
    
    def __iter__(self):
        self._iter_index = 0
        return self
    
    def __next__(self):
        while self._iter_index % self.n_tasks != self.task_id:
            self._iter_index += 1
        if self._iter_index < len(self):
            item = self._batches[self._iter_index]
            self._iter_index += 1
            return item
        else:
            raise StopIteration
            
    def __getitem__(self, index):
        return self.batches[index]

def print_dict_tree(dict_: dict, indent: int=0, indent_symbol: str=" ") -> None:
    """
    Recursively print keys of a dictionary.
    """
    if not isinstance(dict_, dict):
        raise TypeError(f"Please provide a `dict`, type of input is {type(dict_)}")
    for key, values in dict_.items():
        if isinstance(values, dict):
            print(f'{int(indent) * indent_symbol} {key} ' + "{")
            print_dict_tree(values, indent=indent+2, indent_symbol=indent_symbol)
            print(f'{int(indent) * indent_symbol}' + "}")
        elif isinstance(values, list):
            if sum(1 for  _ in values if isinstance(_, dict)) == 0:
                print(f'{int(indent) * indent_symbol} {key} : <value>')
                continue
            print(f'{int(indent) * indent_symbol} {key} [ ')
            has_values = False
            for index, item in enumerate(values):
                if isinstance(item, dict):
                    print(f'{int(indent) * indent_symbol} [{index}] ' + "{")
                    print_dict_tree(item, indent=indent+2, indent_symbol=indent_symbol)
                    print(f'{int(indent) * indent_symbol}' + "}")
                else:
                    if not has_values:
                        print(f'{int(indent+2) * indent_symbol} <value>')
                        has_values = True
                    else:
                        continue
            print(f'{int(indent) * indent_symbol}]')
        else:
            print(f'{int(indent) * indent_symbol} {key} : <value>')
            
def print_key_paths(key_name: str, 
                    dictionary: dict, 
                    parent_path: str="", 
                    child_of: str | None = None) -> None:
    """
    Iterate recursively trough a dict and print all paths that end with a
    given key name.

    **Arguments:**

    key_name: `str`
        key for to search path for

    dictionary: `dict`
        (Nested) dictionary in which the key path is searched.

    **Keyword arguments:**

    parent_path: `str`
        path until current iteration

        default: ""

    child_of: `str` or `None`
        if not `None`, print only paths that contain the string specified here

        default: `None`
    """
    if not isinstance(dictionary, dict):
        raise TypeError(f"Please provide a `dict`, type of input is {type(dictionary)}")
    for key, values in dictionary.items():
        if key == key_name:
            if child_of is not None:
                if child_of in parent_path:
                    print(f"{parent_path}/{key_name}")
            else:
                print(f"{parent_path}/{key_name}")
        if isinstance(values, dict):
            print_key_paths(key_name, values, parent_path=f"{parent_path}/{key}", child_of=child_of)
        elif isinstance(values, list):
            if sum(1 for  _ in values if isinstance(_, dict)) == 0:
                continue
            has_values = False
            for index, item in enumerate(values):
                if isinstance(item, dict):
                    print_key_paths(key_name, item, parent_path=f"{parent_path}/{key}/{index}", child_of=child_of)
                else:
                    if not has_values:
                        has_values = True
                    else:
                        continue    


from abc import ABC, abstractmethod
from typing import Any, List
from .api_core import APIClass
from ..data_framework import Material 
from ..utils import safe_log, tqdm
import ase
import os
import csv
import hashlib
import json

class FileReader(ABC):
    """
    Base class for file reader objects.

    **Keyword arguments:**

    file_path: `str`
        Path to file to be read.

    **Methods:**
    """
    def __init__(self, file_path: str = '.'):
        self.set_file_path(file_path)

    def set_file_path(self, file_path: str) -> None:
        """
        Set the file path.

        **Arguments:**

        file_path: `str`
            File path of file to be read.
        """
        self.file_path = file_path

    @abstractmethod
    def read(file_name, file_path: str = None) -> Any:
        return

class FileReaderASE(FileReader):
    """
    Wrapper class around the ASE IO module.

    Loads geometry from a DFT input file to an ASE Atoms object.

    **Keyword arguments:**

    file_path: `str`
        Path to file to be read.

    **Methods:**
    """

    def read(self, 
             file_name: str, 
             file_path: str = None, 
             **kwargs) -> ase.Atoms:
        """
        Read geometry from DFT input file.

        **Arguments:**

        file_name: `str`
            Name of file to be read. Format is guessed from extension.

        **Keyword arguments:**

        file_path: `str`
            Path of file to be read. Sets the `file_path` attribute.

            default: `None` --> use `self.file_path`

        Additional keyword arguments are forwarded to `ase.io.read`.

        **Returns:**

        atoms: `ase.Atoms`
            Atoms object with geometry read from file.
        """
        if file_path is not None:
            self.set_file_path(file_path)
        atoms = ase.io.read(os.path.join(self.file_path, file_name), **kwargs)
        return atoms

class CSVPropertyReader(FileReader):
    """
    Read material properties from a CSV file and return as dictionary.

    Assumes that the first column of CSV file contains a key. Example content:


    .. code-block:: 

        id,property1,property2   
        0, 1.4, 2.3
        1, 2, 2.5

    Calling `read()` returns:
 
    .. code-block:: python

       {"0" : {"property1" : 1.4, "property2" : 2.3},
       "1" : {"property1" : 2, "property2" : 2.5}}

    **Keyword arguments:**

    file_path: `str`
        Path to file to be read.

    **Methods:**
    """

    def read(self, 
             file_name_csv: str, 
             file_path_csv: str = None, 
             **kwargs) -> dict:
        """
        Read CSV data from file to dictionary.

        **Arguments:**

        file_name_csv: `str`
            Name of CSV file to be read.

        **Keyword arguments:**

        file_path_csv: `str`
            File path to file to be read.

            default: `None` --> use `self.file_path`

        Additional keyword arguments are ignored.

        **Returns:**

        data_dict: `dict`
            Dictionary of properties. First column of csv data is used as dictionary keys.
        """
        if file_path_csv is not None:
            self.set_file_path(file_path_csv)
        with open(os.path.join(self.file_path, file_name_csv),'r') as csv_file:
            reader = csv.reader(csv_file, delimiter=',', quotechar='|')
            data = [x for x in reader]
            header = data[0]
            data = data[1:]
        data_dict = {}
        for entry in data:
            data_dict[entry[0]] = {}
            for key, value in zip(header[1:],entry[1:]):
                data_dict[entry[0]][key] = json.loads(value)
        return data_dict

class API(APIClass):
    """
    API class to add materials from a local folder structure.

    **Requires the following data structure:**

    * root_directory
        * <properties.file>
        * directory[0]
            * <input.file>
        * directory[1]
            * <input.file>
        * ...

    where the directory names correspond to the entries of the first column of the CSV file.

    The name of individual files can be chosen on demand.

    **Keyword arguments**

    root: `str`
        Root directory.

        default: '.'

    file_reader: `FileReader`
        File reader object for reading input files.

        default: `FileReaderASE()`

    property_reader: `FileReader`
        File reader object for reading properties file.

        default: `CSVPropertyReader()`

    logger: `logging.Logger`
        Logger to write error and info messages.

    **Methods:**
    """

    def __init__(self, 
                 root = '.', 
                 file_reader = FileReaderASE(), 
                 property_reader = CSVPropertyReader(), 
                 logger = None):
        self.root = root
        self.log = logger
        self.file_reader = FileReaderASE(file_path = root) if file_reader is None else file_reader
        self.property_reader = CSVPropertyReader(file_path = root) if property_reader is None else property_reader

    def get_calculation(self, 
                        file_path: str, 
                        file_name: str, 
                        property_file_path: str, 
                        property_file_name: str, 
                        property_file_id: str, 
                        file_reader_kwargs: dict = {}, 
                        property_reader_kwargs: dict = {}) -> Material:
        """
        Get calculation from a local file.

        **Arguments:**

        file_path: `str`
            Path to the local file of the structure. Passed to `FileReader.read()` method of file reader. 

        file_name: `str`
            Name of the local file of the structure. Passed to `FileReader.read()` method of file reader.

        property_file_path: `str`
            Path of the file containing the properties of the calculation. 
            Passed to `FileReader.read()` method of property reader.

        property_file_name: `str`
            Name of the file containing the properties of the calculation.
            Passed to `FileReader.read()` method of property reader.

        property_file_id: `str`
            Id of the structure in the property file.
            If no string is provided, it will be converted to string.

        **Keyword arguments:**

        file_reader_kwargs: `dict`
            Additional keyword arguments passed to the file reader object.

            default: `{}`

        property_reader_kwargs: `dict`
            Additional keyword arguments passed to the property reader object.

            default: `{}`

        **Returns:**

        material: `Material`
            Material object with set material id, atomic structure and `data` properties.
        """
        atoms = self.file_reader.read(file_name, file_path = os.path.join(self.root, file_path), **file_reader_kwargs)
        properties = self.property_reader.read(property_file_name, os.path.join(self.root, property_file_path), **property_reader_kwargs)[str(property_file_id)]
        mid = self._gen_mid(file_path, file_name, property_file_path, property_file_name)
        return Material(mid, atoms = atoms, data = properties)

    def get_calculations_by_search(self, 
                                   folder_path: str, 
                                   file_name: str, 
                                   property_file_path: str, 
                                   property_file_name: str, 
                                   file_reader_kwargs: str = {},
                                   show_progress: bool = True, 
                                   property_reader_kwargs: str = {}) -> List[Material]:
        """
        Get calculations from a local file-structure.

        Assumes that every struture is stored in a separate directory, which has a name corresponding to the id extracted from the properties file.

        **Arguments:**

        file_path: `str`
            Path to the local file of the structure. Passed to `FileReader.read()` method of file reader. 

        file_name: `str`
            Name of the local file of the structure. Passed to `FileReader.read()` method of file reader.

        property_file_path: `str`
            Path of the file containing the properties of the calculation. 
            Passed to `FileReader.read()` method of property reader.

        property_file_name: `str`
            Name of the file containing the properties of the calculation.
            Passed to `FileReader.read()` method of property reader.

        **Keyword arguments:**

        file_reader_kwargs: `dict`
            Additional keyword arguments passed to the file reader object.

            default: `{}`

        property_reader_kwargs: `dict`
            Additional keyword arguments passed to the property reader object.

            default: `{}`

        **Returns:**

        materials: `List[Material]`
            List of `Material` objects with set material id, atomic structure and `data` properties.
        """
        calculations = []
        folders = os.listdir(os.path.join(self.root, folder_path))
        for folder in tqdm(folders, disable=not show_progress):
            if not os.path.isdir(os.path.join(self.root, folder_path, folder)):
                safe_log(f"Skipping: {folder} - not a directory", logger=self.log, level="info")
                continue
            try:
                calculations.append(self.get_calculation(os.path.join(folder_path,folder), 
                                                         file_name, 
                                                         property_file_path, 
                                                         property_file_name, 
                                                         property_file_id=folder, 
                                                         file_reader_kwargs=file_reader_kwargs,
                                                         property_reader_kwargs = property_reader_kwargs))
            except Exception as exc:
                error_message = 'Could not read file in folder:' + os.path.join(folder_path,folder) + '! Reason: ' + str(exc)
                safe_log(error_message, logger=self.log)
        return calculations

    def get_property(self, *args, **kwargs):
        """
        Not implemented
        """
        raise NotImplementedError('Not implemented.')

    def _gen_mid(self, *args) -> str:
        """
        Generate mid from arguments by hashing over their `str` values.
        """
        string = ':'.join([str(arg) for arg in args])
        return '!' + hashlib.md5(bytes(string,'utf8')).hexdigest()



from typing import List, Tuple, Callable
import numpy as np
import pandas as pd
import os
import json
import multiprocessing
from functools import partial

from .utils import report_error, BatchIterator
from .fingerprint import Fingerprint

class SimilarityMatrix():
    """
    A matrix, that stores all similarites between materials and the corresponding material identifier.    
    """

    def __init__(self, 
                 matrix: List[list] = [], 
                 mids: List[str] = None, 
                 dtype: type = np.float64):
        '''
        **Keyword arguments:**

        matrix: `np.ndarray` or `List[list]`; 
            Initialize matrix with precomputed similarities.
            
            default: `[]`

        mids: `List[str]`
            Material ids of precomputed similarities.
        
            default: `None` 
        '''
        self._dtype = dtype
        self._dataframe = None
        self.set_matrix(matrix)
        self.set_mids(mids)
        self.fp_type = None
        self.fp_name = None
        self._iter_index = 0

    @property
    def matrix(self) -> np.ndarray:
        """
        Matrix values.
        """
        return self._dataframe.values

    @property
    def mids(self) -> np.ndarray:
        """
        Material identifier, corresponding to rows and columns of the matrix.
        """
        return np.array(self._dataframe.index, dtype=str)

    @property
    def dataframe(self) -> pd.DataFrame:
        """
        Return `pandas.DataFrame` object containing similarities as values and material ids as indices.
        """
        if not hasattr(self, "_dataframe"):
            raise AttributeError("Matrix is not calculated! Can not return data.")
        return self._dataframe.copy()

    def get_metadata(self) -> dict:
        """
        Get dictionary of fingerprint type and name.
        """
        return {"fp_type" : self.fp_type, "fp_name" : self.fp_name}

    def set_metadata(self, metadata: dict) -> None:
        """
        Set fingerprint type and name from dictionary. Ignores other keys than `fp_type` and `fp_name`.
        """
        if "fp_type" in metadata.keys():
            self.fp_type = metadata["fp_type"]
        if "fp_name" in metadata.keys():
            self.fp_name = metadata["fp_name"]

    def set_matrix(self, matrix: np.ndarray) -> None:
        """
        Set the values of the similarity matrix.
        
        **Arguments:**

        matrix: `np.ndarray`
            array of values for the matrix

            Can be a square matrix or an upper triangular matrix.
        """

        if len(matrix) == 0:
            self._dataframe = pd.DataFrame()
            return
        if len(np.unique([len(row) for row in matrix])) != 1:
            matrix = self._square_from_triangular_matrix(matrix)
        if not isinstance(matrix, np.ndarray):
            matrix = np.array(matrix)
        if not isinstance(self._dataframe, pd.DataFrame) or matrix.shape != self._dataframe.shape:
            self._dataframe = pd.DataFrame(data = matrix)
        else:
            self._dataframe.values[:] = np.array(matrix, dtype = self._dtype)

    def set_mids(self, mids: list) -> None:
        """
        Set the matrix row and column identifiers to given values.

        **Arguments:**

        mids: `List[str]`
            List of identifiers to be set
        """
        if isinstance(mids, np.ndarray):
            mids = mids.tolist()
        if mids is None or len(mids) == 0:
            return
        self._dataframe.rename(columns=dict(zip(self._dataframe.columns, mids)), 
                               index=dict(zip(self._dataframe.index, mids)), 
                               inplace=True)

    def set_dataframe(self, dataframe: pd.DataFrame) -> None:
        """
        Set the dataframe of the matrix.

        **Arguments:**

        dataframe: `pandas.DataFrame`
            Dataframe that should contain similarities as values and material identifier as indices.
        """
        self._dataframe = dataframe

    def calculate(self, 
                  fingerprints: List[Fingerprint], 
                  mids: List[str] = None, 
                  multiprocess:  int | None = -1, 
                  symmetric: bool = True):
        """
        Calculate similarity matrix.

        **Arguments:**

        fingerprints: `List[Fingerprint]`
            Fingerprints to calculate similarities.

        **Keyword arguments:**

        mids: `List[str]`
            Material ids of fingerprints.

            default: `None` -> mids are extracted from `fingerprints`

        multiprocess: `int`
            Calculate similarities on available processors.
            Set to `-1` to use all available processors.
            Set to `None` for serial execution
            Set to any positive integer to use as may processes.

            default: `-1`

        symmetric: `bool` 
            Reduce computation cost by calculating only unique half of symmetric matrix
            
            default: `True`

        **Returns:**

        self: `SimilarityMatrix`
            Populated similarity matrix.
        """
        matrix = []
        mids = mids if mids is not None else [fp.mid for fp in fingerprints]
        try:
            self.fp_type = fingerprints[0].fp_type
        except Exception as e:
            report_error(None, f"Unable to set Fingerprint type because of: {str(e)}")
            self.fp_type = "Unknown"
        try:
            self.fp_name = fingerprints[0].name
        except Exception as e:
            report_error(None, f"Unable to set Fingerprint name because of: {str(e)}")
            self.fp_name = "Unknown"
        if multiprocess is not None:
            if multiprocess <= 0:
                if multiprocess == -1:
                    n_processes = multiprocessing.cpu_count()
                else:
                    raise ValueError("Invalid input for `multiprocess`. Choose from positive integer, `-1` or `None`.")
            else:
                n_processes = multiprocess
            if symmetric:
                calc_function = self._get_similarities_list_index
            else:
                calc_function = partial(self._get_similarities_list_index, square_matrix = True)
            with multiprocessing.Pool(n_processes) as p:
                matrix = p.map(calc_function, [(idx, fingerprints) for idx in range(len(fingerprints))]) # TODO refactor, as unneccessary data is copied
                if not symmetric:
                    matrix = np.array(matrix)
        else:
            for idx, fp in enumerate(fingerprints):
                if symmetric:
                    matrix_row = []
                    for fp2 in fingerprints[idx:]:
                        matrix_row.append(fp.get_similarity(fp2))
                else:
                    matrix_row = fp.get_similarities(fingerprints)
                matrix.append(np.array(matrix_row))
        self.set_matrix(matrix)
        self.set_mids(mids)
        return self
    
    def get_sub_matrix(self, 
                       mid_list: List[str], 
                       copy: bool = True) -> object:
        """
        Get sub matrix of all elements in mid_list sorted by occurrence in mid list.

        **Arguments:**

        mid_list: `List[str]`
            List of (unique) mids of materials to include in sub matrix

        **Keyword arguments:**

        copy: `bool`
            Return a new similarity matrix. If set to False, apply changes to `self`.
            
            default: `True` 

        **Returns:**

        `SimilarityMatrix()` object of sub matrix if `copy == True`

        `self` restricted to, and sorted by, elements in `mid_list`
        """
        assert len(set(mid_list)) == len(mid_list), "mid mist contains no-unique entries"
        frame = self.dataframe
        frame = frame[mid_list]
        frame = frame.transpose()
        frame = frame[mid_list]
        new_matrix = frame.values
        if copy:
            simat = SimilarityMatrix(matrix = new_matrix, mids = mid_list, dtype = self._dtype)
            simat.fp_name = self.fp_name
            simat.fp_type = self.fp_type
            return simat
        else:
            self.set_matrix(new_matrix)
            self.set_mids(mid_list)
            return self

    def get_overlap_matrix(self, 
                           column_mids: List[str], 
                           row_mids: List[str]) -> object:
        """
        Get `OverlapSimilarityMatrix()` from matrix. This new matrix contains (mostly) off-diagonal elements of the original matrix.

        **Arguments:**

        row_mids: `List[str]` 
            mids associated with the rows of the new matrix

        column_mids: `List[str]`
            mids associated with the columns of the matrix

        **Returns:**

        overlap_matrix: `OverlapSimilarityMatrix`
            New matrix with rows and columns corresponding to materials with identifier `row_mids` and `column_mids`
        """
        frame = self.dataframe
        frame = frame[column_mids]
        frame = frame.transpose()
        frame = frame[row_mids]
        frame = frame.transpose()
        simat =  OverlapSimilarityMatrix(matrix = frame.values, row_mids = row_mids, column_mids = column_mids)
        simat.fp_type = self.fp_type
        simat.fp_name = self.fp_name
        return simat

    def lookup_similarity(self, 
                          fp1: Fingerprint, 
                          fp2: Fingerprint):
        """
        Return similarity between two fingerprints from the matrix.

        The expected usecase for this is to pass this function `set_similarity_function` of a `Fingerprint` object.

        fp1, fp2: `Fingerprint`
            Fingerprint objects to retrieve similarity

        **Returns:**

        similarity: `float`
            Similarity between materials with mids `fp1.mid` and `fp2.mid`

        **Raises:**

        KeyError: Similarity matrix does not contain an entry with these mids.
        """
        return self.get_entry(fp1.mid, fp2.mid)

    def train_test_split(self, 
                         train_mids: List[str], 
                         test_mids: List[str]) -> set:
        """
        Split similarity matrix into a (symmetric) train matrix and a (off-diagonal) test matrix.

        **Arguments:**

        train_mids: `List[str]`
            mids that identify materials of the training set

        test_mids: `List[str]` 
            mids that identify materials in the test set

        **Returns:**

        train_matrix, test_matrix: `Set[SimilarityMatrix, OverlapSimilarityMatrix]`
        """
        return self.get_sub_matrix(train_mids), self.get_overlap_matrix(train_mids, test_mids) 

    def align(self, matrices): #TODO TEST PROPERLY
        """
        Align the materials in this matrix and all provided matrices.

        **Arguments:**

        matrices: `SimilarityMatrix` or `List[SimilarityMatrix]`
            Matrix or list of matrices to align
        
        **Returns:**

        `None`

        WARNING! Entries in both matrices will be altered, i.e. unique entries in each matrix will be dropped.
        """
        if isinstance(matrices, (list, np.ndarray)):
            #shared_mids = [mid for mid in self.mids if np.array([mid in matrix.mids for matrix in matrices]).all()]
            shared_mids = set.intersection(*[set(simat.mids) for simat in matrices])
            shared_mids = list(shared_mids.intersection(set(self.mids)))
            shared_mids.sort()
            for matrix in matrices:
                matrix.get_sub_matrix(shared_mids, copy = False)
        else:
            shared_mids = [mid for mid in self.mids if mid in matrices.mids]
            matrices.get_sub_matrix(shared_mids, copy = False)
        self.get_sub_matrix(shared_mids, copy = False)

    def get_entry(self, 
                  mid1: str, 
                  mid2: str) -> np.float64:
        """
        Get any entry of the matrix.
        
        **Arguments:**

        mid1, mid2: `str`
            material ids of the requested material
        
        **Returns:**
        
        similarity: `float`
            Similarity between both materials

        **Raises:**

        KeyError: No entry for material with given mid.
        """
        return self._dataframe[mid1][mid2]

    def get_symmetric_matrix(self):
        """
        Deprecated! Use `SimilarityMatrix().matrix` property!
        Get square matrix form.
        
        **Returns:**

        square_matrix: `np.ndarray`
            matrix of similarities
        """
        report_error(None, "Deprecation warning: use SimilarityMatrix.matrix instead of get_symmetric matrix")
        return self.matrix

    def get_row(self, mid: str) -> np.ndarray:
        """
        Get a row of the matrix, by mid or index.

        **Arguments:**

        mid: `str` or `int`
            Material id oder matrix index of requested matrix row

        **Returns:**

        row: `np.ndarray`
            Similarities of material with given mid to all other materials in the matrix.

        **Raises:**

        KeyError: No entry with given mid.

        IndexError: Matrix index out of range.
        """
        if isinstance(mid, int):
            mid = self.mids[mid]
        return np.array(self._dataframe[mid])

    def get_entries(self):
        import warnings
        warnings.warn("Function is deprecated due to ambiguous name. Use `get_unique_entries`.", DeprecationWarning)
        return self.get_unique_entries()

    def get_unique_entries(self) -> np.ndarray:
        """
        Return all enries of the upper triangular matrix.

        **Returns:**

        entries: `np.ndarray`
            list of all unique entries of the matrix
        """
        entries = np.concatenate(self._triangular_from_square_matrix(self.matrix)) #this version should be a lot faster
        return np.array(entries, dtype = np.float64)

    def get_k_most_similar(self, 
                           ref_mid: str, 
                           k: int = 10, 
                           remove_self: bool = True) -> dict:
        """
        Get the k most similar materials and the respective similarities for a material.

        WARNING! Accurate results can only be obtained for symmetric similarity matrices.
        For asymmetric similarity measures, the assignment may be ambiguous.

        **Arguments:**

        ref_mid: `str`
            Material id of the reference material

        **Keyword arguments:**

        k: `int`
            Number of most similar materials to return

            default: 10

        remove_self: `bool`
            Remove the requested material from the results list

            default: `True`

        **Returns:**

        dict: {<1st_nearest_mid>:<similarity>, 2nd_nearest_mid>:<similiarty>, ...}
        """
        row = [(mid,entry) for mid, entry in zip(self.mids, self.get_row(ref_mid))]
        row.sort(reverse = True, key = lambda x: x[1])
        if remove_self:
            for idx, item in enumerate(row):
                if item[0] == ref_mid:
                    row.pop(idx)
                    break
        return {mid: entry for mid, entry in row[:k]}

    def save(self, 
             filename: str = 'similarity_matrix.npy', 
             filepath: str = '.', 
             overwrite: bool = False) -> None:
        """
        Save SimilarityMatrix to numpy binary file.

        **Keyword arguments:**

        filename: `str`
            Name of the file

            default: 'similarity_matrix.npy'

        filepath: `str`
            Relative path to created files
            
            default: '.' 

        overwrite: `bool`
            Overwrite matrix file if it exists.

            default: `False`
        """
        matrix_path = os.path.join(filepath, filename)
        if os.path.exists(matrix_path) and not overwrite:
            raise FileExistsError(f"Matrix at {matrix_path} exists. To overwrite set `overwrite = True`.")
        np.save(matrix_path, np.array([self.mids, self.get_metadata(), self.matrix], dtype = object))

    @staticmethod
    def load(filename: str = 'similarity_matrix.npy', filepath: str = '.') -> object:
        """
        Static method. Load SimilarityMatrix from file. 
        If the target file is created from an OverlapSimilarityMatrix object, return OverlapSimilarityMatrix object

        Warning: This methods loads a pickled file. Only load files of known origin.

        **Keyword arguments:**

        filename: `str` 
            Name of a saved similarity matrix file
            
            default: 'similarity_matrix.npy'

        filepath: `str`
            Relative path to SimilarityMatrix files
            
            default: '.'

        **Returns:**
        
        similarity_matrix: `SimilarityMatrix` or `OverlapSimilarityMatrix`

        **Raises:**

        IndexError: Wrong format of data in file. Can not load.
        """
        matrix_path = os.path.join(filepath, filename)
        matrix = np.load(matrix_path, allow_pickle = True)
        if len(matrix) == 3:
            mids, metadata, matrix = matrix
        elif len(matrix) == 2:
            mids, matrix = matrix
            metadata = {}
        else:
            raise IndexError("Can not load matrix: wrong format of data")
        if len(mids) == 2:
            self = OverlapSimilarityMatrix()
            row_mids = mids[0]
            column_mids = mids[1]
            self.set_matrix(matrix)
            self.set_mids(row_mids, column_mids)
        else:
            self = SimilarityMatrix()
            self.set_matrix(matrix)
            self.set_mids(mids)
        self.set_metadata(metadata)
        return self

    def get_matching_matrices(self, second_matrix):
        """
        Match matrices such that they contain the same materials in the same order.
        
        **Arguments:**

        second_matrix: `SimilarityMatrix`
            Matrix to match materials

        **Returns:**

            new_self, new_matrix: `tuple(SimilarityMatrix)`
                Matching similarity matrices
        """
        shared_mids = [mid for mid in self.mids if mid in second_matrix.mids]
        new_self = self.get_sub_matrix(shared_mids)
        new_matrix = second_matrix.get_sub_matrix(shared_mids)
        return new_self, new_matrix

    def get_cleared_matrix(self, leave_out_mids, copy = True):
        """
        Return a matrix where all materials with mids specified in `leave_out_mids` are removed from the matrix.

        **Arguments:**

        leave_out_mids: `List[str]`
            mids of materials to leave out of the matrix
        
        **Keyword arguments:**

        copy: `bool`
            Return copy of SimilarityMatrix(); apply changes to self, if False
            
            default: `True` 

        **Returns:**

        cleared_matrix: `SimilarityMatrix`
            Copy of similarity matrix if copy is True, `self` otherwise without `leave_out_mids`
        """
        new_mids = [mid for mid in self.mids if mid not in leave_out_mids]
        new_matrix = self.get_sub_matrix(new_mids, copy = copy)
        return new_matrix

    def _get_similarities_list_index(self, idx__list, square_matrix = False): # TODO refactor
        idx, fp_list = idx__list
        if square_matrix:
            sims = np.array(fp_list[idx].get_similarities(fp_list))
        else:
            sims = fp_list[idx].get_similarities(fp_list[idx:])
        return np.array(sims)

    def _triangular_from_square_matrix(self, matrix):
        """
        Get the upper triangular part of a square matrix.
        Args:
            * matrix: list of lists, np.ndarray; symmetric square matrix
        Returns:
            * triangular_matrix: np.ndarray; triangular form of input matrix
        """
        triangular_matrix = []
        for index, row in enumerate(matrix):
            triangular_matrix.append(np.array(row[index:], dtype = self._dtype))
        return np.array(triangular_matrix, dtype = object)

    def _square_from_triangular_matrix(self, matrix):
        symmetric_matrix = np.zeros((len(matrix[0]),len(matrix[0])))
        for idx, row in enumerate(matrix):
            for jdx, entry in enumerate(row):
                symmetric_matrix[idx][jdx + idx] = entry
                symmetric_matrix[jdx + idx][idx] = entry
        return np.array(symmetric_matrix, dtype = self._dtype)

    def __len__(self):
        return len(self._dataframe)

    def __iter__(self):
        self._iter_index = 0
        return self

    def __next__(self):
        if self._iter_index >= len(self):
            self._iter_index = 0
            raise StopIteration
        else:
            self._iter_index += 1
            return self.get_row(self._iter_index - 1)

    def __getitem__(self, key):
        if isinstance(key, str):
            try:
                return self.get_row(key)
            except KeyError:
                raise KeyError("No entry with mid = " + key + '.')
        elif isinstance(key, int):
            try:
                return self._dataframe.values[key]
            except KeyError:
                raise KeyError('No entry with id = ' + str(key) + '.')
        else:
            raise KeyError('Key can not be interpreted as matrix key.')

    def __add__(self, simat):
        new_matrix = SimilarityMatrix()
        new_matrix.set_dataframe(self.dataframe + simat.dataframe)
        return new_matrix

    def __sub__(self, simat):
        new_matrix = SimilarityMatrix()
        new_matrix.set_dataframe(self.dataframe - simat.dataframe)
        return new_matrix

    def __mul__(self, simat):
        new_matrix = SimilarityMatrix()
        if isinstance(simat, SimilarityMatrix):
            new_matrix.set_dataframe(self.dataframe * simat.dataframe)
        else:
            new_matrix.set_dataframe(self.dataframe * simat)
        return new_matrix

    def __eq__(self, simat):
        try:
            return ((self._dataframe == simat._dataframe).all()).all()
        except ValueError:
            sorted_self = self.get_sub_matrix(simat.mids)
            return ((sorted_self._dataframe == simat._dataframe).all()).all()
            
    def __repr__(self) -> str:
        return f"SimilarityMatrix({self.fp_type}, {self.fp_name}, {len(self)})"

class OverlapSimilarityMatrix(SimilarityMatrix):
    """
    A SimilarityMatrix that is used to store similarities between different sets of fingerprints.
    """

    def __init__(self, 
                 matrix: List[list] = [], 
                 row_mids: list = [], 
                 column_mids: list = [], 
                 dtype = np.float64):
        """
        **Keyword arguments:**

        matrix: `List[list]`
            Precomputed matrix values

            default: `[]`

        row_mids: `List[str]`
            Material ids corresponding to the rows of the matrix

            default: `[]`

        column_mids: `List[str]`
            Material ids corresponding to the columns of the matrix

            default: `[]`

        dtype: `type`
            Data type for storing matrix values

            default: numpy.float64``
        """
        self._dtype = dtype
        self._dataframe = None
        self.set_matrix(matrix)
        self.set_mids(row_mids, column_mids)
        self.fp_type = None
        self.fp_name = None
        self._iter_index = 0

    @property
    def mids(self) -> Tuple[List[str], List[str]]:
        """
        Get mids corresponding to matrix row and columns.
        """
        return self.row_mids, self.column_mids

    @property
    def row_mids(self) -> np.ndarray:
        """
        Get mids corresponding to matrix rows.
        """
        return np.array(self._dataframe.index, dtype = str)

    @property
    def column_mids(self) -> np.ndarray:
        """
        Get mids corresponding to matrix columns.
        """
        return np.array(self._dataframe.columns, dtype = str)

    def set_mids(self, 
                 row_mids: List[str], 
                 column_mids: List[str]) -> None:
        """
        Set the mids for rows and columns of the matrix.
        """
        if isinstance(row_mids, np.ndarray):
            row_mids = row_mids.tolist()
        if isinstance(column_mids, np.ndarray):
            column_mids = column_mids.tolist()
        if column_mids is None or len(column_mids) == 0:
            return
        if row_mids is None or len(row_mids) == 0:
            return        
        self._dataframe.rename(columns=dict(zip(self._dataframe.columns, column_mids)), 
                        index=dict(zip(self._dataframe.index, row_mids)), 
                        inplace=True)

    def set_matrix(self, matrix: np.ndarray) -> None:
        """
        Set matrix values

        If the values do not fit the shape of the original matrix, values and mids will be overwritten.
        
        **Arguments:**

        matrix: `np.ndarray`
            Matrix to be set
        """
        if self._dataframe is None:
            self._dataframe = pd.DataFrame(matrix)
        if not isinstance(matrix, np.ndarray):
            matrix = np.array(matrix, dtype=self._dtype)
        if not isinstance(self._dataframe, pd.DataFrame) or matrix.shape != self._dataframe.shape:
            self._dataframe = pd.DataFrame(data = matrix)
        else:
            self._dataframe.values[:] = np.array(matrix, dtype = self._dtype)

    def calculate(self, 
                  reference_fingerprints: List[Fingerprint], 
                  fingerprints: List[Fingerprint], 
                  mids: List[str] = [], 
                  reference_mids = []) -> object:
        """
        Calculate similarity of fingerprints to given reference fingerprints.

        **Arguments:**

        reference_fingerprints: `List[Fingerprint]` 
            Fingerprints that correspond to columns of the matrix
        
        fingerprints: `List[Fingerprint]`
            Fingerprints that correspond to rows of the matrix
        
        **Keyword arguments:**

        mids: `List[str]` 
            Material identifier for the rows of the matrix

            default: `None`

        reference_mids: `List[str]`
            Material identifier for the columns of the matrix
        
            default: `None`

        If possible, mids and reference_mids are taken directly from the Fingerprint objects.

        **Returns:**

        self: `OverlapSimilarityMatrix`
            Calculated matrix object
        """
        try:
            with multiprocessing.Pool() as p:
                matrix = p.map(self._calc_simimilarities_multiprocess, [(fingerprints, reference_fingerprints, idx) for idx in range(len(fingerprints))])
        except TypeError as e: #Fingerprints are not parallelizable
            report_error(None, "Parallel computing not possible, calculating matrix serial.")
            report_error(None, str(e))
            matrix = []
            for fingerprint in fingerprints:
                matrix.append(fingerprint.get_similarities(reference_fingerprints))
        self.set_matrix(np.array([np.array(row) for row in matrix]))
        try:
            row_mids = [fp.mid for fp in fingerprints]
            column_mids = [fp.mid for fp in reference_fingerprints]
            self.set_mids(row_mids, column_mids)
            self.fp_type = fingerprints[0].fp_type
            self.fp_name = fingerprints[0].name
        except AttributeError as e:
            report_error(None, "An error occured, but was caught: ")
            report_error(None, str(e))
            self.set_mids(mids, reference_mids)
        return self

    def get_entries(self) -> np.ndarray:
        """
        Get all entries of the matrix as a list.

        **Returns:**

        entries: `numpy.ndarray`
            All entries of the matrix in a (N*M, 1)-dim list
        """
        return np.array(self.matrix).flatten()

    def get_row(self, mid: str | int) -> np.ndarray:
        """
        Get a row of the matrix.

        **Arguments:**

        mid: `str` or `int`
            Material id or matrix index of the requested row

        **Returns:**

        matrix_row: `numpy.ndarray`
            Row of the matrix
        """
        if isinstance(mid, int):
            mid = self.row_mids[mid]
        return np.array(self.dataframe[mid:mid])[0]

    def get_column(self, mid:  str | int) -> np.ndarray:
        """
        Get a column of the matrix.

        **Arguments:**

        mid: `str` or `int`
            Material id or matrix index of the requested column

        **Returns:**

        matrix_column: `numpy.ndarray`
            Column of the matrix
        """
        if isinstance(mid, int):
            mid = self.column_mids[mid]
        return np.array(self.dataframe[mid])

    def get_symmetric_matrix(self):
        """
        Not implemented for OverlapSimilarityMatrix().

        This function has no meaning for overlap similarity matrices.

        **Raises:**

        NotImplementedError: upon function call
        """
        raise NotImplementedError("OverlapSimilarityMatrix() does not support symmetric matrices.")

    def get_sub_matrix(self, 
                       row_mid_list: List[str], 
                       column_mid_list: List[str], 
                       copy: bool = True) -> object:
        """
        Get sub matrix.

        **Arguments:**

        column_mid_list: `List[str]` 
            list of mids of materials in matrix column to include in sub matrix
            
        row_mid_list: `List[str]` 
            list of mids of materials in matrix row to include in sub matrix
        
        **Keyword arguments:**

        copy: `bool`
            Return a new similarity matrix. If set to False, apply changes to ``self``.

            default: `True`

        **Returns:**

        matrix: `OverlapSimilarityMatrix` 
            if `copy == True`
            
        self: `OverlapSimilarityMatrix` 
            if `copy == False`, self, restricted to, and sorted by, elements in `mid_list`
        """
        frame = self.dataframe
        frame = frame[column_mid_list]
        frame = frame.transpose()
        frame = frame[row_mid_list]
        frame = frame.transpose()
        if copy:
            new_matix = OverlapSimilarityMatrix()
            new_matix.set_dataframe(frame)
            return new_matix #OverlapSimilarityMatrix(matrix = frame.values, column_mids = column_mid_list, row_mids = row_mid_list)
        else:
            self.set_matrix(frame.values)
            self.set_mids(row_mid_list, column_mid_list)
            return self

    def save(self, 
             filename: str = 'overlap_similarity_matrix.npy', 
             filepath: str = '.', 
             overwrite: bool = False) -> None:
        """
        Save SimilarityMatrix to numpy binary file(s).

        **Keyword arguments:**

        filename: `str`
            name of the matrix file

            default: 'similarity_matrix.npy'
        
        data_path: `str`
            relative path to created files

            default: '.' 
        
        overwrite: `bool`
            Overwrite matrix file if it exists.

            default: `False`
        """
        matrix_path = os.path.join(filepath, filename)
        if os.path.exists(matrix_path) and not overwrite:
            raise FileExistsError(f"Matrix at {matrix_path} exists. To overwrite set `overwrite = True`.")
        np.save(matrix_path, np.array([[self.row_mids, self.column_mids], self.matrix], dtype=object))

    def get_entry(self, 
                  row_mid: str,
                  column_mid: str) -> np.float64:
        """
        Get a single entry of the matrix.

        **Arguments:**
    
        row_mid: `str`
            material id of the material in the row of the matrix
        
        column_mid: `str`
            material id of the material in the column of the matrix
        
        **Returns:**
        
        similarity: `numpy.float64`
            similarity between both materials
        """
        return self._dataframe[column_mid][row_mid]

    def transpose(self) -> None:
        """
        Exchange rows and columns of matrix.
        """
        self._dataframe = self.dataframe.transpose()

    def _calc_simimilarities_multiprocess(self, tfp__kfp__idx):
        target_fingerprints, kernel_fingerprints, idx = tfp__kfp__idx
        fps = target_fingerprints[idx].get_similarities(kernel_fingerprints)
        return fps

    def align(self, matrices: List[SimilarityMatrix] | SimilarityMatrix) -> None:
        """
        Align the materials in this matrix and all provided matrices.

        **Arguments:**

        matrices: `OverlapSimilarityMatrix` or `List[OverlapSimilarityMatrix]`
            matrix object(s) to align with
        
        **Returns:**
        
        `None`

        Warning! Entries in both matrices will be altered, i.e. unique entries in each matrix will be dropped.
        """
        if isinstance(matrices, (list, np.ndarray)):
            shared_row_mids = [mid for mid in self.row_mids if np.array([mid in matrix.row_mids for matrix in matrices]).all()]
            shared_column_mids = [mid for mid in self.column_mids if np.array([mid in matrix.column_mids for matrix in matrices]).all()]
            for matrix in matrices:
                matrix.get_sub_matrix(shared_row_mids, shared_column_mids, copy = False) 
        else:
            shared_row_mids = [mid for mid in self.row_mids if mid in matrices.row_mids]
            shared_column_mids = [mid for mid in self.column_mids if mid in matrices.column_mids]
            matrices.get_sub_matrix(shared_row_mids, shared_column_mids, copy = False)
        self.get_sub_matrix(shared_row_mids, shared_column_mids, copy = False)

    def _check_matrix_alignment(self, simat):
        if (self.column_mids == simat.column_mids).all() and (self.row_mids == simat.row_mids).all():
            return True
        return False

    def __add__(self, simat):
        if isinstance(simat, OverlapSimilarityMatrix):
            if not self._check_matrix_alignment(simat):
                raise IndexError("Matrices not aligned. Use ``align()`` method.")
            matrix = self.matrix + simat.matrix
        elif np.isscalar(simat):
            matrix = simat + self.matrix
        else:
            raise TypeError("Unknown variable type.")
        return OverlapSimilarityMatrix(matrix, row_mids = self.row_mids, column_mids = self.column_mids)

    def __sub__(self, simat):
        if isinstance(simat, OverlapSimilarityMatrix):
            if not self._check_matrix_alignment(simat):
                raise IndexError("Matrices not aligned. Use ``align()`` method.")
            matrix = self.matrix - simat.matrix
        elif np.isscalar(simat):
            matrix = simat - self.matrix
        else:
            raise TypeError("Unknown variable type.")
        return OverlapSimilarityMatrix(matrix, row_mids = self.row_mids, column_mids = self.column_mids)

    def __mul__(self, simat):
        if isinstance(simat, OverlapSimilarityMatrix):
            if not self._check_matrix_alignment(simat):
                raise IndexError("Matrices not aligned. Use ``align()`` method.")
            matrix = self.matrix * simat.matrix
        elif np.isscalar(simat):
            matrix = simat * self.matrix
        else:
            raise TypeError("Unknown variable type.")
        return OverlapSimilarityMatrix(matrix, row_mids = self.row_mids, column_mids = self.column_mids)

    def __eq__(self, simat):
        try:
            return ((self._dataframe == simat._dataframe).all()).all()
        except ValueError:
            sorted_self = self.get_sub_matrix(simat.row_mids, simat.column_mids)
            return ((sorted_self._dataframe == simat._dataframe).all()).all()            

class BatchedSimilarityMatrix():
    """
    A similarity matrix for parallel computation distributed over different tasks.

    The calculation of pairwise similarities between fingerprints can be distributed over
    different processors. To do so, the matrix is split into different independent 
    sub-matrices (batches), which are calculated separately.

    Each `BatchedSimilarityMatrix` stores all required data and metadata in a single folder.

    To reduce memory consumption, the fingerprints are input via a separate file containing 
    serialized `Fingerprint` objects. This file can be generated via, _e.g._:

    .. code-block:: python

        import json

        long_fingerprint_list_data = [fp.serialize() for fp in long_fingerprint_list]

        with open('path/to/long/fingerprint/file', 'w') as fingerprint_file:
            json.dump(long_fingerprint_list_data, fingerprint_file)

    In a next step, a `BatchedSimilarityMatrix` object can be used to split this large file
    into separate batches for computation of the similarity matrix.

    .. code-block:: python

        batched_similarity_matrix.fingerprint_file_batches('long_fingerprint_file_name',
                                                           'path/to/long/fingerprint/file')

    Now all data is prepared and the matrix can be calculated:

    .. code-block:: python
    
        batched_similarity_matrix.calculate(similarity_function)

    Note that because the fingerprints were serialized before, the similarity function must be 
    specified for calculating similarities.

        **Keyword arguments:**

        root_path: `str`
            Path where the data folder shall be created.

        default: "."

        matrix_folder_name: `str`
            Name of the similarity matrix data folder.

            This name should be unique and descriptive!

            default: "batched_similarity_matrix"

        fingerprint_files_name: `str`
            Base name of fingerprint files to be generated.

        load_from_file: `bool`
            Load metadata from file and ignore further keyword arguments.

            default: `True`

        batch_size: `int`
            Maximal number of fingerprints in a batch. 
            A similarity matrix has a maximum of `batch_size` ** 2 entries.

            default:  `10000`

        n_tasks: `int`
            Total number of tasks that are used to compute the similarity matrix.

            default: `1`

        task_id: `int`
            Id of the current task. 
            This number specifies which batches of the similarity matrix are calculated.

            default: `0`

        size: `int`
            Total number of fingerprints. 
            This number is automatically set if the function `fingerprint_file_batches` is called.

            default: `0`

        symmetric: `bool`
            Assume that similarity matrix is symmetric and calculate only unique batches.
            Setting this option to `True` reduces the number of batches that are calculated 
            by ca. a factor of two, as off-diagonal elements are computed only once.

    **Methods:**
    """

    def __init__(self, 
                 root_path: str = ".", 
                 matrix_folder_name: str = "batched_similarity_matrix",
                 fingerprint_files_name: str = "batch_similarity_fingerprints", 
                 load_from_file: bool = True,
                 batch_size: int = 10000,
                 n_tasks: int = 1, 
                 task_id: int = 0,
                 size: int = 0,
                 symmetric: bool = True,
                 dtype = np.float32):
        folder_path = os.path.join(root_path, matrix_folder_name)
        if not os.path.exists(folder_path):
            os.makedirs(folder_path)
        self.matrix_folder_name = matrix_folder_name
        self._folder_path = os.path.abspath(folder_path)
        self._task_id = task_id
        self._dtype = dtype
        if load_from_file and os.path.exists(os.path.join(self.folder_path, self.metadata_filename)):
            metadata = self._metadata_from_file()
            self.fingerprint_files_name = metadata["fingerprint_files_name"]
            self._n_tasks = metadata["n_tasks"]
            self._batch_size = metadata["batch_size"]
            self._size = metadata["size"]
            self._symmetric = metadata["symmetric"]
        else:
            self.fingerprint_files_name = str(fingerprint_files_name)
            self._n_tasks = int(n_tasks)
            self._batch_size = int(batch_size)
            self._size = int(size)
            self._symmetric = bool(symmetric)
        
    @property
    def n_tasks(self) -> int:
        """
        Total number of tasks used to calculate similarity matrix.
        """
        return self._n_tasks
    
    @property
    def task_id(self) -> int:
        """
        Task id of the current task.
        """
        return self._task_id
    
    @property
    def batch_size(self) -> int:
        """
        Number of fingerprints in a batch.
        """
        return self._batch_size
    
    @property
    def folder_path(self) -> str:
        """
        Path to folder where all data corresponding to this matrix is stored.
        """
        return self._folder_path

    @property
    def size(self) -> int:
        """
        Total number of fingerprints that are used.
        """
        return self._size

    @property
    def symmetric(self) -> bool:
        """
        Matrix is symmetric.
        """
        return self._symmetric
    
    @property
    def batch_iterator(self) -> BatchIterator:
        """
        `BatchIterator` object that is used to iterate over batches of the similarity matrix.
        """
        return BatchIterator(self.size, self.batch_size, self.n_tasks, self.task_id, self.symmetric)
    
    @property
    def metadata(self) -> dict:
        """
        Metadata for BatchedSimilarityMatrix.
        """
        metadata = {
            "fingerprint_files_name" : self.fingerprint_files_name,
            "folder_path" : self.folder_path,
            "n_tasks" : self.n_tasks,
            "batch_size" : self.batch_size,
            "size" : self.size,
            "symmetric" : self.symmetric
        }
        return metadata
    
    @property
    def metadata_filename(self) -> str:
        """
        Name of the metadata file.
        """
        return f"{self.matrix_folder_name}_metadata.json"

    @property
    def mid_filename(self) -> str:
        """
        Name of the file where all material ids are stored.
        """
        return f"{self.matrix_folder_name}_mids.txt"
    
    @property
    def most_similar_materials_filename(self) -> str:
        """
        (Base) name of file(s) where all most similar materials are stored.
        """
        return f"{self.matrix_folder_name}_most_similar_materials_{self.task_id}.txt"
    
    @property
    def matrices_for_this_task(self):
        """
        Return the number of similarity matrices that are calculated for this task.
        """
        n_tasks = len([_batch for _batch in self.batch_iterator])
        return n_tasks

    def set_task_id(self, task_id: int) -> None:
        assert task_id < self.n_tasks, f"task id must be smaller than number of tasks {self.n_tasks}"
        self._task_id = task_id

    def set_n_tasks(self, n_tasks: int, write_to_metadata: bool = False):
        assert self.task_id < n_tasks, f"current task id is smaller than number of tasks {n_tasks}"
        self._n_tasks = n_tasks
        if write_to_metadata:
            self.save_metadata()

    def save_metadata(self, overwrite: bool = True) -> None:
        """
        Save current metadata to `self.folder_name/self.metadata_filename`.

        **Keyword arguments:**

        overwrite: `bool`
            Overwrite file if it exists.

            default: `True`
        """
        filename = os.path.join(self.folder_path, self.metadata_filename)
        if os.path.exists(filename) and not overwrite:
            raise FileExistsError("Metadata file exists! Use `overwrite = True` to overwrite.")
        with open(filename, "w") as metadata_file:
            json.dump(self.metadata, metadata_file)
        
    def fingerprint_file_batches(self, 
                                 fingerprint_file_name: str, 
                                 fingerprint_file_path: str,
                                 overwrite: bool = False,
                                 write_mid_file: bool = True,
                                 save_metadata_updates: bool = True) -> None:
        """
        Split a single, large fingerprint file into batches that can be read by the individual
        tasks of a `BatchedSimilarityMatrix`. 

        **Arguments:**

        fingerprint_file_name: `str`
            Name of the file containing a json-encoded list of serialized `Fingerprint` objects.

        fingerprint_file_path: `str`
            Path to fingerprint file.

        **Keyword arguments:**

        overwrite: `bool`
            Overwrite batched fingerprint files if they exist.
            If this is set to `False` and (any of) the files exist(s), a `FileExistsError` is raised.

            default: `False`

        write_mid_file: `bool`
            Write a file that contains an enumeration of all material ids in the similarity matrix.

            default: `True`

        save_metadata_updates: `bool`
            Save the metadata after determining the total number of fingerprints.

            default: `True`
        """
        with open(os.path.join(fingerprint_file_path, fingerprint_file_name), "r") as fp_file:
            file_data = json.load(fp_file)
        self._size = len(file_data)
        if save_metadata_updates:
            self.save_metadata()
        fingerprint_batches = BatchIterator.linear_batch_list(self.size, self.batch_size)
        for idx, jdx in fingerprint_batches:
            filename = self.gen_fingerprint_batch_file_name(idx, jdx)
            filepath = os.path.join(self.folder_path, filename)
            if os.path.exists(filepath) and not overwrite:
                raise FileExistsError("Fingerprint file exists! Use `overwrite = True` to overwrite.")
            with open(filepath, "w") as fp_file:
                json.dump(file_data[idx:jdx], fp_file)
        if write_mid_file:
            filepath = os.path.join(self.folder_path, self.mid_filename)
            if os.path.exists(filepath) and not overwrite:
                raise FileExistsError("Mid file exists! Use `overwrite = True` to overwrite.")
            with open(filepath, "w") as mid_file:
                for idx, data_entry in enumerate(file_data):
                    mid_file.write(f"{idx} {json.loads(data_entry)['mid']}\n")
        
    def gen_fingerprint_batch_file_name(self, idx: int, jdx: int) -> str:
        """
        Generate name of a fingerprint batch file from indices.

        **Arguments:**

        idx, jdx: `int`
            Start and end index of the fingerprints file batch.

        **Returns:**

        filename: `str`
            Name of a file with given indices.
        """
        return f"{self.fingerprint_files_name}_{idx}_{jdx}.json"
    
    def gen_matrix_batch_file_name(self, batch: List[List[int]]) -> str:
        """
        Generate name of a matrix batch file from batch.

        **Arguments:**

        batch: `List[List[int]]`
            batch, as returned by `BatchIterator` objects

        **Returns:**

        filename: `str`
            Name of a file corresponding to given batch.
        """
        return f"{self.matrix_folder_name}_{batch[0][0]}_{batch[0][1]}__{batch[1][0]}_{batch[1][1]}.npy"
        
    def fingerprints_from_batch_file(self, idx: int, jdx: int) -> List[Fingerprint]:
        """
        Read fingerprints from file and deserialize. Generates a list of Fingerprints from the files written by
        the function `fingerprint_file_batches`.

        idx, jdx: `int`
            Start and end index of the fingerprints file batch.

        **Returns:**

        fingerprints: `List[Fingerprint]`
            Fingerprint objects from file
        """
        filename = self.gen_fingerprint_batch_file_name(idx, jdx)
        filepath = os.path.join(self.folder_path, filename)
        with open(filepath, "r") as fp_file:
            fps = json.load(fp_file)
        return [Fingerprint.deserialize(fp) for fp in fps]
    
    def get_row_by_mid(self, mid: str) -> List[float]:
        """
        Get a row of the (already calculated) similarity matrix from the mid of a material.

        Calling this function requires the the mid file was written by `fingerprint_file_batches`.

        **Arguments:**

        mid: `str`
            Material id of the reference entry.

        **Returns:**

        similarities: `List[float]`
            Similarities of all materials of this similarity matrix to the reference specified by `mid`

        **Raises:**

        `FileNotFoundError`: Mid file does not exist (or has a non-compatible name).

        `KeyError`: No material of given `mid` in the mid file (and thus in the matrix).
        """
        matrix_row = []
        mid_file_path = os.path.join(self.folder_path, self.mid_filename)
        if not os.path.exists(mid_file_path):
            raise FileNotFoundError("Mid file not found. Can not extract row by mid.")
        # linear search through file to find index of mid in matrix
        ref_idx = None
        with open(mid_file_path, "r") as mid_file:
            for line in mid_file:
                idx, line_mid = line.strip().split()
                if line_mid == mid:
                    ref_idx = int(idx)
                    break
        if ref_idx is None:
            raise KeyError("Mid was not found in file!")
        # open matrices that contain this entry
        for batch in self.batch_iterator.get_batches_for_index(ref_idx):
            matrix_name = self.gen_matrix_batch_file_name(batch)
            matrix = SimilarityMatrix.load(filename = matrix_name, filepath=self.folder_path)
            if batch[1][0] <= ref_idx < batch[1][1]:
                matrix_row.extend(matrix[mid])
            else:
                matrix_row.extend(matrix.get_column(mid))
        return matrix_row
            
    def calculate(self, 
                  similarity_function: Callable, 
                  overwrite: bool = False):
        """
        Calculate similarity matrix entries and write them to files.

        **Arguments:**

        similarity_function: `Callable`
            Function of two `Fingerprint` objects that returns their similarity.

        **Keyword arguments:**
        
        overwrite: `bool`
            Recalculate and overwrite existing matrix batch files. 
        """
        for batch in self.batch_iterator:
            filename = self.gen_matrix_batch_file_name(batch)
            if os.path.exists(os.path.join(self.folder_path, filename)) and not overwrite:
                print(f"Matrix {filename} exists. Skipping.")
                # ignore existing files
                continue
            fingerprints = self.fingerprints_from_batch_file(batch[0][0], batch[0][1])
            for fp in fingerprints:
                fp.set_similarity_function(similarity_function)
            if batch[0] == batch[1]: # diagonal elements
                matrix = SimilarityMatrix(dtype=self._dtype).calculate(fingerprints, symmetric=self.symmetric)
            else:
                row_fingerprints = self.fingerprints_from_batch_file(batch[1][0], batch[1][1])
                for fp in row_fingerprints:
                    fp.set_similarity_function(similarity_function)
                matrix = OverlapSimilarityMatrix(dtype=self._dtype).calculate(fingerprints, row_fingerprints)
            matrix.save(filename = filename, filepath = self.folder_path, overwrite = overwrite)
    
    def write_most_similar_materials_file(self, 
                                          k: int = 10, 
                                          remove_self: bool = True):
        """
        Write files containing all `k` most similar materials for each entry of the matrix.

        **Keyword arguments:**

        k: `int`
            Number of most similar materials to find.

            default: `10`

        remove_self: `bool`
            Find most similar materials *without* self.

            default: `True`
        """
        self.k = int(k) 
        for batch_row in self.batch_iterator.get_batch_rows():
            if remove_self:
                self.k += 1
            matrix_position = -1
            position_list = []
            for batch in batch_row:
                if batch[0] == batch[1]:
                    position_list.append(0)
                    matrix_position = 1
                    continue
                position_list.append(matrix_position)
            with multiprocessing.Pool() as pool:
                most_similar_batches = pool.map(self._get_most_similar_batch, zip(batch_row, position_list))
            if remove_self:
                self.k -= 1
            most_similar = self._merge_most_similar_batches(most_similar_batches, remove_self = remove_self)
            self._append_most_similar_materials_file(most_similar)

    def get_entry_histogram(self, bins = np.arange(0,1.01,0.01)):
        """
        Generate a histogram of entries in the matrix.

        **Keyword arguments:**

        bins: `np.ndarray`
            Bins of the histogram. Each bin contains the number of matrix entries 
            in the range [bins[i], bins[i+1]).

            default: `np.arange(0,1.01,0.01)`

        **Returns:**

        (entries, bins): `(np.ndarray, np.ndarray)`
            entries[i] contains the absolute number of matrix entries with similarity
            in the range [bins[i], bins[i+1]).
        """
        entries = np.zeros(len(bins)-1)
        for batch in self.batch_iterator.batches:
            filename = self.gen_matrix_batch_file_name(batch)
            simat = SimilarityMatrix.load(filename = filename, filepath=self.folder_path)
            if type(simat) == SimilarityMatrix:
                in_matrix_entries = simat.get_unique_entries()
            elif type(simat) == OverlapSimilarityMatrix:
                in_matrix_entries = simat.get_entries()
            hist_, _ = np.histogram(in_matrix_entries, bins = bins)
            entries += hist_
        return entries, bins
    
    def _get_most_similar_batch(self, batch_position):
        batch, position = batch_position
        most_similar_batch = []
        matrix_file_name = self.gen_matrix_batch_file_name(batch)
        matrix = SimilarityMatrix.load(filename = matrix_file_name, filepath = self.folder_path)
        # unify matrix format
        if position == 0: # diagonals
            mids = matrix.mids
            ref_mids = mids
        elif position == -1: # horizontal
            mids = matrix.column_mids
            ref_mids = matrix.row_mids
        else: # vertical
            mids = matrix.row_mids
            ref_mids = matrix.column_mids
            matrix.transpose()
        matrix = matrix.matrix.tolist()
        # add for each matrix row the most similar materials
        for row in matrix:
            most_similar_batch.append(sorted(zip(row, mids), reverse = True)[:self.k])
        return ref_mids, most_similar_batch
    
    def _merge_most_similar_batches(self, most_similar_batches, remove_self = True):
        ref_mids = []
        most_similar = []
        for batch in most_similar_batches:
            ref_mids.append(batch[0])
            most_similar.append(np.array(batch[1]))
        for idx in range(len(ref_mids)-1):
            assert all(ref_mids[idx] == ref_mids[idx+1]), "Reference mids do not coincide."
        ref_mids = ref_mids[0]
        full_list = np.concatenate(most_similar, axis = 1).tolist()
        best_per_row = []
        for ref_mid, sub_list in zip(ref_mids, full_list):
            if remove_self:
                sub_list = filter(lambda x: x[1] != ref_mid, sub_list)
            best_per_row.append(sorted(sub_list, reverse=True)[:self.k])
        return ref_mids, best_per_row

    def _append_most_similar_materials_file(self, most_similar):
        most_similar_materials_filepath = os.path.join(self.folder_path, self.most_similar_materials_filename)
        with open(most_similar_materials_filepath, "a") as outfile:
            for ref_mid, most_similar_materials in zip(*most_similar):
                data = {ref_mid : {mid: sim for sim,mid in most_similar_materials}}
                outfile.write(json.dumps(data))
                outfile.write("\n")
    
    def _metadata_from_file(self):
        filename = os.path.join(self.folder_path, self.metadata_filename)
        with open(filename, "r") as metadata_file:
            metadata = json.load(metadata_file)
        return metadata

    def __repr__(self) -> str:
        return f"BatchedSimilarityMatrix(size = {self.size}) at {self.folder_path}"


from typing import Callable, List, Set
import numpy as np
from functools import partial, cache

class NotCreatedError(Exception):
    """
    Empty container for naming the error occuring when the Grid was initialized, but not created.
    """
    pass

class Grid():

    """
    Grid object used to discretize DOS spectra for the generation of DOS fingerprints.
    To obtain a Grid() object, create it with:
        grid = Grid.create(**parameters)    
    """
    
    @staticmethod
    def create(grid_id = None,
               grid_type = "nonuniform", 
               e_ref = -2,
               delta_e_min = 0.05, 
               delta_e_max = 1.05, 
               delta_rho_min = 0.5, 
               delta_rho_max = 5.5, 
               width = 7, 
               n_pix = 56, 
               cutoff = [-8,7],
               energy_discretization = None,
               states_discretization = None):
        """
        Create grid object. 

        There are two options to initialize the grid: 
            1. Initialize from grid id: The grid id, which is a string that contains all grid parameters, is used to set the parameters of the grid.
            2. Initialize from parameters: The grid is created directly from the parameters passed as key-word arguments.

        **Parameters**

        grid_id: *str* or *None*
            Identifier to re-create a grid, or None, if new or default parameters are given.

        grid_type: *str*
            Currently two options are available:
                * *nonuniform* - Create a non-uniform grid from parameters
                * *uniform* - Create an uniform grid from parameters
                * *custom* - Create a custom grid from lists defining the limits of the grid
            
        e_ref: float
            Reference energy in eV, i.e. the center of the grid.

        delta_e_min: float
            Minimal energy difference in eV between two grid points.

        delta_e_max: float
            Maximal energy difference in eV between two grid points.
            Ignored if *uniform* grid is created.
        
        delta_rho_min: float
            Minimal difference of states in an energy interval.
        
        delta_rho_max: float
            Maximal difference of states in an energy interval.
            Ignored if *uniform* grid is created.
        
        width: float
            Width of the feature region of the Grid in eV.
            Ignored if *uniform* grid is created.

        n_pix: int
            Number of intervals to discretize states in an energy interval.
        
        cutoff: (float, float)
            Lowest and highes energies in eV to be included in the grid.

        energy_discretization: list[float]
            Used by *custom* grid - ignored else.
            Values of energy intervals of the grid, in eV.

        states_discretization: list[float]
            Used by *custom* grid - ignored else.
            Maximal number of states for each energy interval of the grid.       

        **Returns**

        self: *Grid*
            The grid object with set grid parameters.

        **Raises**

        TypeError:
            Grid type can not be identified (neither "nonuniform", nor "uniform", nor "custom").

        ValueError:
            Grid type can no be identified.
        """
        if grid_id is None:
            self = Grid()
            self.grid_type = grid_type
            self.e_ref = e_ref
            self.delta_e_min = delta_e_min
            self.delta_e_max = delta_e_max
            self.delta_rho_min = delta_rho_min
            self.delta_rho_max = delta_rho_max
            self.width = width
            self.cutoff_min = cutoff[0]
            self.cutoff_max = cutoff[1]
            self.n_pix = n_pix
            self.energy_discretization = energy_discretization
            self.states_discretization = states_discretization
            self.grid_id = self.get_grid_id()
        else:
            self = Grid.create(**Grid.resolve_grid_id(grid_id))
        return self

    def get_grid_id(self):
        self._check_created()
        if self.grid_type == "nonuniform":
            grid_id_values = [self.grid_type, self.e_ref, self.delta_e_min, self.delta_e_max, self.delta_rho_min, self.delta_rho_max, self.width, self.cutoff_min, self.cutoff_max, self.n_pix]
        elif self.grid_type == "uniform":
            grid_id_values = [self.grid_type, self.e_ref, self.delta_e_min, self.delta_rho_min, self.cutoff_min, self.cutoff_max, self.n_pix]
        elif self.grid_type == "custom":
            if self.energy_discretization is None:
                raise ValueError("Must provide valid energy discretization for custom fingerprint.")
            if self.states_discretization is None:
                raise ValueError("Must provide valid states discretization for custom fingerprint.")
            from hashlib import md5
            all_values = [self.n_pix, self.e_ref]
            all_values = np.append(all_values, self.energy_discretization)
            all_values = np.append(all_values, self.states_discretization)
            all_values = "".join([str(value) for value in all_values])
            return f"custom:{md5(all_values.encode()).hexdigest()}"
        else:
            raise ValueError("Grid type can not be identified")
        grid_id = ':'.join([str(value) for value in grid_id_values])
        return grid_id

    @staticmethod
    def resolve_grid_id(grid_id: str) -> dict:
        grid_variables = {}
        grid_id_variables = grid_id.split(':')
        if grid_id_variables[0] == "nonuniform":
            grid_variables["grid_type"] = grid_id_variables[0] 
            grid_variables["e_ref"] = float(grid_id_variables[1])
            grid_variables["delta_e_min"] = float(grid_id_variables[2])
            grid_variables["delta_e_max"] = float(grid_id_variables[3])
            grid_variables["delta_rho_min"] = float(grid_id_variables[4])
            grid_variables["delta_rho_max"] = float(grid_id_variables[5])
            grid_variables["width"] = float(grid_id_variables[6])
            grid_variables["cutoff"] = [float(grid_id_variables[7])]
            grid_variables["cutoff"].append(float(grid_id_variables[8]))
            grid_variables["n_pix"] = int(grid_id_variables[9])
        elif grid_id_variables[0] == "uniform":
            grid_variables["grid_type"] = grid_id_variables[0]
            grid_variables["e_ref"] = float(grid_id_variables[1])
            grid_variables["delta_e_min"] = float(grid_id_variables[2])
            grid_variables["delta_rho_min"] = float(grid_id_variables[3])
            grid_variables["cutoff"] = [float(grid_id_variables[4])]
            grid_variables["cutoff"].append(float(grid_id_variables[5]))
            grid_variables["n_pix"] = int(grid_id_variables[6])
        elif grid_id_variables[0] == "custom":
            raise TypeError("Custom grid id can not be resolved.")
        else:
            raise ValueError("Grid type can not be identified")
        return grid_variables

    def grid(self) -> list:
        """
        Generate a grid with parameters set by `self.grid_id`.

        **Returns**

        grid: *list*
            Nested list of type [[<grid_point_energy_i>, [<dos_bins_i>]], ...]
        """
        self._check_created()
        if self.grid_type == "nonuniform":
            grid = self._grid_non_uniform(self.delta_e_min, self.delta_e_max, self.delta_rho_min, self.delta_rho_max, self.width, self.n_pix, (self.cutoff_min, self.cutoff_max))
        elif self.grid_type == "uniform":
            grid = self._grid_uniform(self.delta_e_min, self.delta_rho_min, self.n_pix, (self.cutoff_min, self.cutoff_max))
        elif self.grid_type == "custom":
            grid = self._grid_custom(tuple(self.energy_discretization), tuple(self.states_discretization), self.n_pix)
        else:
            raise ValueError("Grid type can not be identified")
        return grid

    def copy(self) -> object:
        """
        Create a copy of self.
        """
        return Grid.create(grid_id=self.get_grid_id())

    @cache
    def _grid_non_uniform(self, delta_e_min, delta_e_max, delta_rho_min, delta_rho_max, width, n_pix, cutoff):
        """
        New implementation of the grid. Follows the description in the publication.
        """

        f_energies = partial(self.gauss_function, w_min = delta_e_min, w_max = delta_e_max, sigma = width) 

        energies = self.energy_intervals_from_function(f_energies, delta_e_min, cutoff)

        f_heights = partial(self.gauss_function, w_min = delta_rho_min, w_max = delta_rho_max, sigma = width) 

        max_heights = self.grid_height_from_function(f_heights, energies, delta_rho_min)

        grid = self.grid_from_lists(energies, max_heights, n_pix)

        return grid

    @cache
    def _grid_uniform(self, delta_e, delta_rho, n_pix, cutoff):

        def f_energies(x):
            return 1

        energies = self.energy_intervals_from_function(f_energies, delta_e, cutoff)

        max_heights = self.grid_height_from_function(f_energies, energies, delta_rho)

        grid = self.grid_from_lists(energies, max_heights, n_pix)

        return grid
        
    @cache
    def _grid_custom(self, energy_discretization, states_discretization, n_pix):
        return self.grid_from_lists(energy_discretization, states_discretization, n_pix)

    def grid_from_lists(self, energies: List[float], max_heights: List[float], n_bins: int) -> List[List[float]]:
        """
        Generate a grid from lists of energies, maximal heights, and the number of bins.

        **Arguments**

        energies: *list*
            List of energy boundaries, aligned at the negative edge of the energy interval for energy bin.

        max_heights: *list*
            List of maximal heights, defined for each energy boundary in `energies`.

        n_bins: *int*
            Number of bins that is used to discretise each energy interval.
        
        **Returns**

        grid: *list*
            Nested list of type [[<grid_point_energy_i>, [<dos_bin_ij>, ...]], ...]
        """
        # avoid unexpected behaviour
        assert len(energies) == len(max_heights), "Number of energy intervals is not equal to the number of maximal heights for each energy."

        # generate grid
        grid = []
        for energy, max_height in zip(energies, max_heights):
            bin_height = max_height / n_bins
            grid.append([energy, [round(idx * bin_height, 10) for idx in range(1, n_bins + 1)]])
        return grid

    def get_grid_indices_for_energy_range(self, energy: List[float]) -> Set[int]:
        """
        Compute the grid indices that match a given energy range.

        **Arguments**

            energy: *list*
            Energies to match grid energies.

        **Returns**

            grid_start, grid_end: set
            First and last index of the grid that are in the limits of the energies.
        """
        grid_energies = [x[0] for x in self.grid()]
        energy = [e for e in energy if (e >= grid_energies[0] and e <= grid_energies[-1])]
        for idx, grid_e in enumerate(grid_energies):
            if grid_e >= energy[0]:
                grid_start = idx if idx >= 0 else 0
                break
        for idx, grid_e in reversed(list(enumerate(grid_energies))):
            if grid_e <= energy[-1]:
                grid_end = idx  
                break
        return grid_start, grid_end

    def energy_intervals_from_function(self, function: Callable, minimal_interval: float, energy_limits: List[float]) -> list:
        """
        Generate a set of energy intervals from a given function. The energy intervals are generated as:

        e_i = \\sum_{j=0}^{i-1} * minimal_interval * function(e_j)

        and 

        e_{-i} = -1 * e_i

        **Arguments**

        function: *Callable*
            function that maps an energy to a number greater or equal 1

        minimal_interval: *float*
            minimal interval width, i.e. minimal distance between two intervals

        energy_limits: *List[float]*
            list [<minimal_energy>, <maximal_energy>] that determines the limits between which energy intervals should be calculated
        """
        # assumes the spectrum is shifted to the reference energy, which is then located at e = 0
        energies = [0]
        
        # we need the maximum absolute limit to generate intervals
        max_limit = max([abs(lim_) for lim_ in energy_limits])
        
        while energies[-1] < max_limit:
            
            # make sure the minimal interval is kept
            next_step = max([function(energies[-1]), 1])
            
            # get new interval width
            next_step *= minimal_interval

            # apply to series
            next_step += energies[-1]

            # increase numerical stability
            next_step = np.round(next_step, 8)

            # update
            if next_step <= max_limit:
                energies.append(next_step)
            else:
                break
        
        # generate negative indices
        full_set = [-1 * energy for energy in energies[1:] if energy <= abs(energy_limits[0])]
        
        # filter positive energies to match limits, and merge negative and positive indices
        full_set.extend(list(filter(lambda x: x <= energy_limits[1], energies)))

        # make sure the sorting is correct
        full_set.sort()

        return full_set

    def grid_height_from_function(self, function: Callable, energies: list, minimal_height: float) -> list:
        """
        Use provided function to calculate the grid height at each energy.

        **Arguments**

        function: *Callable*
            Function that assigns an integer number to an energy value

        energies: *list*
            List of energies for which the bin heights are calculated

        minimal_height: *float*
            Minimal height of a bin
        """
        return [max([function(energy), 1]) * minimal_height for energy in energies]

    def gauss_function(self, x: float, w_min: float, w_max: float, sigma: float) -> int:
        """
        Function that assigns an integer value to each `x`.

        **Arguments**

        x: *float*
            Input value

        w_min: *float*
            Minimal width of the interval

        w_max: *float*
            Maximal width of the interval

        sigma: *float*
            Parameter controlling the width of the interal gaussian function
        """
        g = (1 - np.exp(-0.5 * (x / sigma)**2))
        value = g * (w_max / w_min - 1) + 1
        return int(value)

    def _check_created(self):
        if not hasattr(self, "grid_type"):
            raise NotCreatedError("Grid was not created. To do so, use Grid.create(**parameters).")
        
    def __repr__(self):
        return f"Grid({self.get_grid_id()})"

    def __eq__(self, other):
        if not self.grid_type == other.grid_type:
            return False
        if self.grid_type == 'nonuniform':
            grid_attributes = [
                "e_ref",
                "delta_e_min",
                "delta_e_max",
                "delta_rho_min",
                "delta_rho_max",
                "width",
                "cutoff_min",
                "cutoff_max",
                "n_pix"
            ]
        elif self.grid_type == 'uniform':
            grid_attributes = [
                "e_ref",
                "delta_e_min",
                "delta_rho_min",
                "cutoff_min",
                "cutoff_max",
                "n_pix"
            ]
        else:
            grid_attributes = [
                "states_discretization",
                "energy_discretization",
                "n_pix"
            ]
        for attribute in grid_attributes:
            if not getattr(self, attribute) == getattr(other, attribute):
                return False
        return True

    def __hash__(self):
        """Define the hash for caching calls to grid.grid(). 
        Grids with the same grid_id should be identical, therefore their hash should be a sufficient hash.
        Differences because 'e_ref = 2' and 'e_ref = 2.0' lead to different hashes, but the overhead should 
        be not too large."""
        return hash(self.get_grid_id())


import numpy as np

from madas.fingerprints import DOSFingerprint

def test_generation():
    x = np.linspace(-10,5,100)
    y = np.sin(x)
    grid = DOSFingerprint.get_default_grid()
    grid.e_ref=0
    grid.cutoff_min = -10
    grid.cutoff_max = 5
    grid.grid_type = "uniform"
    fp = DOSFingerprint(grid_id=grid.get_grid_id()).calculate(x, y, convert_data=None)
    assert fp.get_grid() == grid, "did not create correct grid from grid_id"
    assert fp.get_similarity(fp) == 1.0, "Similarity to self is not 1"

    # repeat test for creation from Grid object directly
    fp = DOSFingerprint(grid=grid).calculate(x, y, convert_data=None)
    assert fp.get_grid() == grid, "did not set grid"
    assert fp.get_similarity(fp) == 1.0, "Similarity to self is not 1"


import json

import pytest
import numpy as np

from madas.utils import BatchIterator, print_dict_tree, print_key_paths, JSONNumpyEncoder

@pytest.fixture
def example_dict():
    return {
        "a" : {
            "b" : 2, 
            "c" : [
                {"d" : 3},
                {"e" : 4},
                5,
                "f"
            ]
        },
        "g" : [1, 2, 3]
    }


def test_BatchIterator_init():

    with pytest.raises(ValueError):
        _ = BatchIterator(4, 2, task_id=1)

def test_BatchIterator_get_batches_for_index():
    
    bi = BatchIterator(11, 3)

    ref_index = 3

    batches_with_ref_index = [
        [[0, 3], [3, 6]], 
        [[3, 6], [3, 6]], 
        [[3, 6], [6, 9]], 
        [[3, 6], [9, 11]], 
    ]

    assert bi.get_batches_for_index(ref_index) == batches_with_ref_index, "Did not return correct batches for index"

    bi = BatchIterator(11, 3, symmetric=False)

    ref_index = 3

    batches_with_ref_index = [
        [[0, 3], [3, 6]], 
        [[3, 6], [3, 6]], 
        [[6, 9], [3, 6]], 
        [[9, 11], [3, 6]], 
    ]

    assert bi.get_batches_for_index(ref_index) == batches_with_ref_index, "Did not return correct batches for index for asymmetric case"

def test_BatchIterator_get_batch_rows():

    bi_t1 = BatchIterator(8,3, n_tasks=2, task_id=0)
    bi_t2 = BatchIterator(8,3, n_tasks=2, task_id=1)

    rows_t1 = [
        [
            [[0, 3], [0, 3]], 
            [[0, 3], [3, 6]], 
            [[0, 3], [6, 8]]
        ], 
        [
            [[0, 3], [6, 8]], 
            [[3, 6], [6, 8]], 
            [[6, 8], [6, 8]]]
    ]

    rows_t2 = [
        [
            [[0, 3], [3, 6]], 
            [[3, 6], [3, 6]], 
            [[3, 6], [6, 8]]
        ]
    ]

    assert bi_t1.get_batch_rows() == rows_t1, "Wrong batch rows for task 0"
    assert bi_t2.get_batch_rows() == rows_t2, "Wrong batch rows for task 1"

def test_BatchIterator_iter():

    size = 11
    n_tasks = 3
    batch_size = 2
    for task_id in range(n_tasks):
        bi = BatchIterator(size,batch_size,task_id=task_id, n_tasks = n_tasks, symmetric=True)
        batches_split = list(bi)
        assert batches_split == bi.batches[task_id::n_tasks], f"Task {task_id} did not return correct batches"

def test_BatchIterator_linear_batch_list():

    linear_list = BatchIterator.linear_batch_list(5,2)

    expected_list = [
        [0,2],
        [2,4],
        [4,5]
    ]

    assert linear_list == expected_list, "Wrong linear list"

def test_print_dict_tree(example_dict, capsys):
    print_dict_tree(example_dict)
    out, err = capsys.readouterr()
    assert len(err) == 0, f"Produced an error: {err}"
    assert out == """ a {
   b : <value>
   c [ 
   [0] {
     d : <value>
  }
   [1] {
     e : <value>
  }
     <value>
  ]
}
 g : <value>
""", "Did not print correct tree!"
    
def test_print_key_paths(example_dict, capsys):
    print_key_paths("e", example_dict)
    out, err = capsys.readouterr()
    assert len(err) == 0, f"Produced an error: {err}"
    assert out == """/a/c/1/e\n""", "Did not print correct path!"

def test_JSONNumpyEncoder():

    assert json.loads(json.dumps([np.single(2), np.longdouble(0.3), np.array([1,2,3,4], dtype=np.int64)], 
                                 cls=JSONNumpyEncoder))==[2,0.3,[1,2,3,4]], "Failed to encode and decode values"


from typing import List
from .backend_core import Backend
from madas.data_framework import Material
from madas.utils import safe_log
from ase.db import connect

class ASEBackend(Backend):
    """
    Database backend that connects to a ASEAtomsDatabase object.
    """
    
    def __init__(self, 
                 filename = "materials_database.db", 
                 filepath = "data", 
                 make_dirs = True,
                 key_name = "mid", 
                 log = None) -> None:
        super().__init__(filename, filepath, make_dirs, key_name, log)
        self._db = connect(self.abs_path)
        with self._db:
            self._metadata = self._db.metadata
        self._write_buffer = {}

    def add_single(self, material: Material) -> None:
        """
        Add a single material to the database.

        **Arguments:**

        material: *madas.Material*
            Material object to be added to the database

        **Return:**

        `None`
        """
        self._db.write(material.atoms, **{self.key_name : material.mid}, data = material.data, **material.properties)
        safe_log(self._log_write_message(material.mid), logger=self.log, level = "info")

    def add_many(self, materials: List[Material]) -> None:
        """
        Add many materials to the database.

        **Arguments:**

        materials: *List[madas.Material]*
            List of material objects to be added to the database

        **Return:**

        `None`
        """
        with self._db as db:
            for material in materials:
                db.write(material.atoms, **{self.key_name : material.mid}, data = material.data, **material.properties)
                safe_log(self._log_write_message(material.mid), logger=self.log, level = "info")
            
    def get_single(self, mid: str = None, **kwargs) -> Material:
        """
        Retrieve a single entry from the database.

        **Keyword arguments**

        mid: *str*
            default: `None`

            Material identifier used in the database and madas.Material objects

        Additional kwargs are used to retrieve entries from the database, e.g. if a single Material object has a specific property value. 
        
        **Returns:**

        material: *madas.Material*
            Material object of the selected database entry

        **Raises:**

        *ValueError*
            If neither a mid nor a keyword argument is given, this error is raised.

        *AssertionError*
            If more than one row matches the keyword arguments, the ASE AtomsDatabase raises this error.

        *KeyError*
            If no material has the specified mid or property values, this error is raised.
        """
        self._check_select_arguments(mid, **kwargs)
        row = self._get_single_row(mid, **kwargs)
        properties = row.key_value_pairs
        mid = properties.pop(self.key_name)
        material = Material(mid, row.toatoms(), data = row.data, properties = properties)
        return material

    def get_many(self, mids: List[str] = None, **kwargs) -> List[Material]:
        """
        Retrieve many entries from the database.

        **Keyword arguments**

        mids: *List[str]*
            default: `None`

            List of material identifiers used in the database and madas.Material objects

        Additional kwargs are used to retrieve entries from the database, e.g. if several Material objects have a specific property value. 
        
        **Returns:**

        materials: *List[madas.Material]*
            List of Material objects of the selected database entries

        **Raises:**

        *ValueError*
            If neither a mid nor a keyword argument is given, this error is raised.

        *KeyError*
            If no material has the specified mid or property values, this error is raised.
        """
        self._check_select_arguments(mids, **kwargs)
        materials = []
        if mids is not  None:
            for mid in mids:
                row = self._db.get(**{self.key_name : mid})
                properties = row.key_value_pairs
                properties.pop(self.key_name)
                materials.append(Material(mid, row.toatoms(), data = row.data, properties = properties))
        else:
            rows = self._db.select(**kwargs)
            for row in rows:
                properties = row.key_value_pairs
                mid = properties.pop(self.key_name)
                materials.append(Material(mid, row.toatoms(), data = row.data, properties = properties))
        return materials

    def get_by_id(self, db_id):
        """
        Get a single entry by the integer valued database id, i.e. the count of when the material was added.

        WARNING: Different than the ase.AtomsDatabase, this index starts with entry 0. 

        **Arguments**

        db_id: *int*
            Index of material in database

        **Returns**

        material: *madas.Material*
            Material object associated with the database entry

        *KeyError*
            If no material has the specified index, this error is raised.
        """
        row = self._db.get(db_id+1)
        properties = row.key_value_pairs
        mid = properties.pop(self.key_name)
        return Material(mid, row.toatoms(), row.data, properties)

    def update_single(self, mid: str = None, **kwargs) -> None:
        """
        Update a single entry from the database.

        **Keyword arguments**

        mid: *str*
            default: `None`

            Material identifier used in the database and madas.Material objects

        Additional kwargs are used to identify entries from the database, e.g. if a single Material object has a specific property value. 
        
        **Returns:**

        `None`

        **Raises:**

        *ValueError*
            If neither a mid nor a keyword argument is given, this error is raised.

        *AssertionError*
            If more than one row matches the keyword arguments, the ASE AtomsDatabase raises this error.

        *KeyError*
            If no material has the specified mid or property values, this error is raised.
        """
        self._check_select_arguments(mid, **kwargs)
        row = self._get_single_row(mid, **kwargs)
        _id = row.id
        self._db.update(_id, **kwargs)

    def update_many(self, mids: List[str] = None, kwargs_list: List[dict] = []) -> None:
        """
        Update a set of entries from the database.

        **Keyword arguments**

        mids: *List[str]*
            default: `None`

            List of material identifiers used in the database and madas.Material objects
        
        kwargs_list: `List[dict]`
            default: `[]`

            List of dictionaries that contain the data that should be written as properties to the database.

        **Returns:**

        `None`

        **Raises:**

        *ValueError*
            If neither a mid nor a keyword argument is given, this error is raised.

        *AssertionError*
            If more than one row matches the keyword arguments, the ASE AtomsDatabase raises this error.

        *KeyError*
            If no material has the specified mid or property values, this error is raised.
        """
        buffer_size = 1000
        ids = []
        for mid in mids:
            self._check_select_arguments(mid)
            row = self._get_single_row(mid)
            ids.append(row.id)
        update_buffer = []
        for update_single in zip(ids, kwargs_list):
            if len(update_buffer) >= buffer_size:
                self._update_buffer(update_buffer)
                update_buffer = []
            update_buffer.append(update_single)
        if len(update_buffer) > 0:
            self._update_buffer(update_buffer)

    def _update_buffer(self, buffer):
        with self._db:
            for _id, kwargs in buffer:
                self._db.update(_id, **kwargs)


    def update_metadata(self, **kwargs) -> None:
        """
        Update database metadata.
        
        Keyword arguments are directly written to database metadata.

        **Returns:**

        `None`
        """
        metadata = self.metadata
        metadata.update(**kwargs)
        self._db.metadata = metadata
        self._metadata = metadata

    def has_entry(self, entry_id: str) -> bool:
        """
        Test if an entry with given id exists in the database.

        **Arguments:**

        entry_id: *str*
            Material identifier of the expected entry.

        **Returns:**

        has_entry: *bool*
            True if an entry with this key exists, False else
        """
        try:
            self._db.get(**{self.key_name : entry_id})
            return True
        except KeyError:
            return False        

    def get_length(self):
        """
        Return the length of the database, i.e. the number of entries.
        """
        return self._db.count()

    def _get_single_row(self, mid: str, **kwargs) -> object:
        if mid is not None:
            row = self._db.get(**{self.key_name : mid})
        else:
            row = self._db.get(**kwargs)
            mid = row.__getattribute__(self.key_name)
        return row



import pytest
from madas.clustering import SimilarityMatrixClusterer
from madas import SimilarityMatrix, Fingerprint

from sklearn.datasets import make_blobs

@pytest.fixture()
def blobs_and_labels():
    return make_blobs(random_state=1)

@pytest.fixture()
def fingerprints(blobs_and_labels):
    pos, _ = blobs_and_labels
    return [Fingerprint("DUMMY").from_material(".".join(map(str, pos_)), value=pos_) for pos_ in pos]

@pytest.fixture()
def similarity_matrix(fingerprints):
    simat = SimilarityMatrix().calculate(fingerprints)
    return simat

@pytest.fixture()
def fitted_clusterer(similarity_matrix):
    simat = similarity_matrix
    clus = SimilarityMatrixClusterer(simat)
    clus.set_clusterer_params(eps = 0.65)
    clus.cluster()
    return clus

def test_fit(similarity_matrix, blobs_and_labels):
    _, labels = blobs_and_labels
    simat = similarity_matrix
    clus = SimilarityMatrixClusterer(simat)

    assert (clus.matrix == 1 - simat.matrix).all(), "Distance matrix is not complement of similarity matrix"
    assert (clus.mids == simat.mids).all(), "Did not copy correct mids"

    clus.set_clusterer_params(eps = 0.65)
    clus.cluster()

    assert len(set(labels)) == clus.nclusters, "Did not recover correct number of clusters"
    assert (labels == clus.labels).all(), "Cluster labels are not correct"

def test_save_and_load(fitted_clusterer, tmpdir):
    fitted_clusterer.save(filepath = tmpdir)
    loaded_clusterer = SimilarityMatrixClusterer.load(filepath = tmpdir)
    
    assert (fitted_clusterer.labels == loaded_clusterer.labels).all(), "Loading did not return same cluster labels"
    assert fitted_clusterer.simat == loaded_clusterer.simat, "Loading did not return same SimilarityMatrix"
    assert fitted_clusterer.use_complement == loaded_clusterer.use_complement, "Loading did not return correct complement handling"

def test_get_mids_sorted_by_cluster_labels(fitted_clusterer):
    label_dict = fitted_clusterer.get_label_dict()
    current_label = min(list(label_dict.values()))
    for mid in fitted_clusterer.get_mids_sorted_by_cluster_labels():
        lab_ = label_dict[mid]
        assert lab_ >= current_label, "Later label larger than precessedor"
        if lab_ > current_label:
            current_label = lab_



from typing import Union, Iterable

import matplotlib.pyplot as plt

from nomad_dos_fingerprints import DOSFingerprint, Grid

def _apply_grid_offset(grid_array: list, offset: float):
    for idx in range(len(grid_array)):
        grid_array[idx][0] += offset
    return grid_array

def plot_fingerprint_in_grid(fingerprint: DOSFingerprint, 
                             show: bool = True, 
                             label: str = '', 
                             axes: Union[plt.Axes, None] = None, **kwargs) -> None:
    """
    Plot a fingerprint in its grid representation.

    **Arguments:**

    fingerprint: `nomad_dos_fingerprints.DOSFingerprint`
        Fingerprint to be plotted

    **Keyword arguments:**

    show: `bool`
        Show the plot.

        default = `True`
    
    label: `str`
        Label for the plot. Labels are required for legends in plots.

        default: `''`

    axes: `matplotlib.pyplot.Axes` or `None`
        Axes to draw fingerprint on. If `None`, create new axes.

        default: `None`
    
    Additional keyword arguments are passed to matplotlib.pyplot.bar(*, **kwargs).
    """
    x, y, all_width = [], [], []
    bin_fp=fingerprint.get_bitarray()
    grid_indices=fingerprint.indices
    grid = Grid.create(grid_id=fingerprint.grid_id)
    plotgrid=_apply_grid_offset(grid.grid(), grid.e_ref)
    plotgrid=plotgrid[grid_indices[0]:grid_indices[1]]

    bit_position=0
    for index,item in enumerate(plotgrid):
        if index<len(plotgrid)-1:
            width=plotgrid[index+1][0]-item[0]
        else:
            width=abs(item[0]-plotgrid[index-1][0])
        for idx, dos_value in enumerate(item[1]):
            if bin_fp[bit_position]==1 and (idx == len(item[1])-1 or bin_fp[bit_position+1] == 0 ):
                x.append(item[0])
                y.append(dos_value)
                all_width.append(width)
            bit_position+=1
    if axes is None:
        plt.bar(x,y,width=all_width,align='edge', label = label, **kwargs)
    else:
        axes.bar(x,y,width=all_width,align='edge', label = label, **kwargs)
    if show:
        plt.show()

def plot_horizontal_lines(grid: Grid, 
                          plot_style: dict = {'c':'k'}, 
                          axes: Union[plt.Axes, None] = None, 
                          only_top: bool = False) -> None:
    """
    Plot horizontal lines of a grid.

    **Arguments:**

    grid: `nomad_dos_fingerprints.Grid`
        Grid to plot.

    **Keyword arguments:**
    
    plot_style: `dict`
        Style of grid lines

        default: `{'c' : 'k'}` 

    axes: `matplotlib.pyplot.Axes` or `None`
        Axes to draw on. If `None`, create new axes.

        default: `None`

    only_top: `bool`
        Show only the highest line

        default: `False`
    """
    grid_array = _apply_grid_offset(grid.grid(), grid.e_ref)
    if axes is None:
        axes = plt.gca()
    e_, r_ = grid_array[0]
    for idx, entry in enumerate(grid_array[1:]):
        e, rhos = entry
        if list(rhos) == list(r_):
            continue
        elif rhos[-1] > r_[-1]: # increasing bin height
            if only_top:
                axes.plot([e_, e], [max(r_), max(r_)], **plot_style)
            else:
                for rho in r_:
                    axes.plot([e_, e], [rho, rho], **plot_style)
            e_, r_ = entry
        elif rhos[-1] < r_[-1]: # decreasing bin height
            if only_top:
                axes.plot([e_, grid_array[idx][0]], [max(r_), max(r_)], **plot_style)
            else:
                for rho in r_:
                    axes.plot([e_, grid_array[idx][0]], [rho, rho], **plot_style)
            e_, r_ = grid_array[idx][0], rhos # energy of the *last* bin
    if only_top:
        axes.plot([e_, e], [max(r_), max(r_)], **plot_style)
    else:
        for rho in r_:
            axes.plot([e_, e], [rho, rho], **plot_style)


def plot_vertical_lines(grid: Grid, 
                        plot_style: dict = {'c':'k'}, 
                        axes: Union[plt.Axes, None] = None) -> None:
    """
    Plot horizontal lines of a grid.

    **Arguments:**

    grid: `nomad_dos_fingerprints.Grid`
        Grid to plot.

    **Keyword arguments:**
    
    plot_style: `dict`
        Style of grid lines

        default: `{'c' : 'k'}` 

    axes: `matplotlib.pyplot.Axes` or `None`
        Axes to draw on. If `None`, create new axes.

        default: `None`
    """
    grid_array = _apply_grid_offset(grid.grid(), offset=grid.e_ref)
    if axes is None:
        axes = plt.gca()
    for e, rhos in grid_array:
        axes.plot([e,e], [0,max(rhos)], **plot_style)

def plot_grid(grid: Grid, 
              vertical: bool = True, 
              horizontal: bool = True, 
              horizontal_only_top: bool = False, 
              plot_style: dict = {'c':'k', 'linewidth' : 0.5}, 
              figure: bool = True, 
              figsize: tuple = (15,10), 
              limits: Union[Iterable, None] = None, 
              show: bool = True, 
              axes:  Union[plt.Axes, None] = None) -> None:
    """
    Plot a Grid.

    **Arguments:**

    grid: `nomad_dos_fingerprints.Grid`
        Grid to plot

    **Keyword arguments:**

    vertical: `bool`
        Plot vertical grid lines

        default: `True`

    horizontal: `bool`
        Plot horizontal grid lines

        default: `True`
    
    horizontal_only_top: `bool`
        Plot only uppermost horizontal grid lines

        default: `False`

    plot_style: `dict`
        Style of grid lines

        default: `{'c' : 'k'}` 

    figure: `bool`
        Create figure

        default: `True`

    figsize: `tuple`
        Size of create figure,
        ignored if `figure` is `False`.

        default: `(15,10)`
    
    limits: `list` of `float` or `None`
        if not `None`, contains the axis limits:
        `[x_min, x_max, y_min, y_max]`

        default: `None`

    show: `bool`
        Show the plot.

        default = `True`

    axes: `matplotlib.pyplot.Axes` or `None`
        Axes to draw on. If `None`, create new axes.

        default: `None`

    """
    if figure:
        plt.figure(figsize=figsize)
    if axes is None:
        axes = plt.gca()
    if vertical:
        plot_vertical_lines(grid, plot_style=plot_style, axes = axes)
    if horizontal:
        plot_horizontal_lines(grid, plot_style=plot_style, axes = axes, only_top=horizontal_only_top)
    if limits is not None:
        axes.set_xlim(*limits[:2])
        axes.set_ylim(*limits[2:])
    if show:
        plt.show()



import pytest

from madas.backend.dictionary_backend import DictBackend
from madas import Material

from ase.build import bulk
from ase import Atoms

@pytest.fixture()
def backend():
    _backend = DictBackend()
    return _backend

@pytest.fixture()
def material():
    return Material("Si:test", atoms = bulk("Si"), data = {"test" : "data"}, properties={"test" : "me"})

@pytest.fixture()
def materials():
    return [Material(f"Si:test{idx}", atoms = bulk("Si"), data = {"test" : f"data{idx}"}, properties={"test" : "me"}) for idx in range(1000)]

def test_add_single(backend, material, capsys):
    backend.add_single(material)
    captured = capsys.readouterr()
    assert Atoms.fromdict(backend._dict["Si:test"]["atoms"]) == material.atoms, "Wrong atoms are written"
    assert captured.out.strip() == backend._log_write_message("Si:test"), "Logging to STDOUT failed."

def test_get_single(backend, material):
    backend.add_single(material)
    mat = backend.get_single(mid = material.mid)
    assert material == mat, "Did not return correct (same) material from database"

def test_add_many(backend, materials):
    backend.add_many(materials)
    for ref in materials:
        mat = backend.get_single(mid = ref.mid)
        assert ref == mat, f"Wrong data for entry with mid {ref.mid}"

def test_get_many(backend, materials):
    backend.add_many(materials)
    mid_list = [mat.mid for mat in materials]
    mats = backend.get_many(mids = mid_list)
    for mat in mats:
        assert mat in materials, "Got material that was not in initial list"
    for mat in materials:
        assert mat in mats, "Material was not returned by database"

def test_get_by_id(backend, material):
    backend.add_single(material)
    mat = backend.get_by_id(0)
    assert material == mat, "Did not return correct material from database"

def test_update_single(backend, material):
    backend.add_single(material)
    backend.update_single(material.mid, something = "new")
    backend.update_single(material.mid, different = 1)
    mat = backend.get_single(material.mid)
    assert mat.properties["something"] == "new", "Failed to update property of single material"
    assert mat.properties["different"] == 1, "Failed to update integer property of single material"

def test_update_many(backend, materials):
    backend.add_many(materials)
    mids = [mat.mid for mat in materials]
    backend.update_many(mids, [{"something" : f"different{idx}"} for idx in range(len(materials))])
    mats = backend.get_many(mids)
    for idx, mat in enumerate(mats):
        print(mat)
        assert mat.properties["something"] == f"different{idx}", f"Failed to update property of many materials: {mat}"

def test_metadata():
    backend = DictBackend()
    backend.update_metadata(something = "new")
    assert backend.metadata == {"something" : "new"}, "Failed to update metadata"


def test_has_entry(backend, material):
    backend.add_single(material)
    assert not backend.has_entry("a"), "Non existing entry reported"
    assert backend.has_entry(material.mid), "Exisiting entry not found"

def test_get_length(backend, material):
    assert backend.get_length() == 0, "Finite length for empty database"
    backend.add_single(material)
    assert backend.get_length() == 1, "Wrong length for one-entry database"

def test_to_json(backend, material):
    backend.add_single(material)
    data = backend.to_json()
    restored = DictBackend.from_json(data)
    assert backend.get_single(material.mid) == restored.get_single(material.mid), "Serialized database did not contain same data"


from importlib.metadata import version

__version__ = version('madas')

# Commonly used classes
from .material import Material  # noqa: F401
from .data_framework import MaterialsDatabase  # noqa: F401
from .similarity import SimilarityMatrix  # noqa: F401
from .fingerprint import Fingerprint  # noqa: F401



from madas import Fingerprint

from ase import Atoms
import numpy as np

def PTE_similarity(fingerprint1, fingerprint2):
    """
    Calculate similarity between PTE fingerprints.

    The similarity is calculated as:

    *similarity = 1 / ( 1 + |PTE1 - PTE2|)*
    """
    diff = abs(fingerprint1.data["averaged_PTE_column"] - fingerprint2.data["averaged_PTE_column"])
    return 1 / (1 + diff)

class PTEFingerprint(Fingerprint):
    """
    A fingerprint comparing the column of the periodic table of elements for all species of the material.
    """
    def __init__(self, 
                 name = "PTE", 
                 similarity_function=PTE_similarity, 
                 pass_on_exceptions=True) -> None:
        self.set_fp_type("PTE")
        self.set_name(name)
        self.set_pass_on_exceptions(pass_on_exceptions)
        self.set_similarity_function(similarity_function)

    def from_material(self, material):
        """
        Calculate PTEFingerprint from a `MADAS` `Material` object.
        """
        self.set_mid(material)
        return self.calculate(material.atoms)
    
    def calculate(self, atoms: Atoms) -> object | None:
        """
        Calculate average periodic table of elements column number.
        """
        column_numbers = []
        for number in atoms.get_atomic_numbers():
                column_numbers.append(COLUMN_PER_ATOM[number])
        self.set_data("averaged_PTE_column", np.mean(column_numbers))
        return self


MAIN_GROUP_DATA = {
    1 : [1,3,11,19,37, 55, 87],
    2 : [2,4,12,20,38,56,88],
    3 : [21,39],
    4 : [22,40,72,104],
    5 : [23,41,73,105],
    6 : [24,42,74,106],
    7 : [25,43,75,107],
    8 : [26,44,76,108],
    9 : [27,45,77,109],
    10 : [28, 46, 78, 110],
    11 : [29, 47, 79, 111],
    12 : [30, 48, 80, 112],
    13 : [5, 13, 31, 49, 81, 113],
    14 : [6, 14, 32, 50, 82, 114],
    15 : [7, 15, 33, 51, 83, 115],
    16 : [8, 16, 34, 52, 84, 116],
    17 : [9, 17, 35, 53, 85, 117],
    18 : [10, 18, 36, 54, 86, 118]
}

MAIN_GROUP_DATA[3].extend(range(57,72)) # Lanthanides
MAIN_GROUP_DATA[3].extend(range(89,104)) # Actinides

COLUMN_PER_ATOM = {}
for group, atomic_numbers in MAIN_GROUP_DATA.items():
    for number in atomic_numbers:
        COLUMN_PER_ATOM[number] = group